#!/usr/bin/env python3
# coding: utf-8

# ###############################################################################
# This file is part of PanACOTA.                                                #
#                                                                               #
# Authors: Amandine Perrin                                                      #
# Copyright © 2018-2020 Institut Pasteur (Paris).                               #
# See the COPYRIGHT file for details.                                           #
#                                                                               #
# PanACOTA is a software providing tools for large scale bacterial comparative  #
# genomics. From a set of complete and/or draft genomes, you can:               #
#    -  Do a quality control of your strains, to eliminate poor quality         #
# genomes, which would not give any information for the comparative study       #
#    -  Uniformly annotate all genomes                                          #
#    -  Do a Pan-genome                                                         #
#    -  Do a Core or Persistent genome                                          #
#    -  Align all Core/Persistent families                                      #
#    -  Infer a phylogenetic tree from the Core/Persistent families             #
#                                                                               #
# PanACOTA is free software: you can redistribute it and/or modify it under the #
# terms of the Affero GNU General Public License as published by the Free       #
# Software Foundation, either version 3 of the License, or (at your option)     #
# any later version.                                                            #
#                                                                               #
# PanACOTA is distributed in the hope that it will be useful, but WITHOUT ANY   #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS     #
# FOR A PARTICULAR PURPOSE. See the Affero GNU General Public License           #
# for more details.                                                             #
#                                                                               #
# You should have received a copy of the Affero GNU General Public License      #
# along with PanACOTA (COPYING file).                                           #
# If not, see <https://www.gnu.org/licenses/>.                                  #
# ###############################################################################

"""
Functions to generate a persistent genome from a pangenome.

@author gem
April 2017
"""
import logging
import math

from PanACoTA import utils

logger = logging.getLogger("corepers.pers")


def get_pers(fam_by_strain, fam_all_members, nb_strains, tol=1, multi=False, mixed=False,
             floor=False):
    """
    From the list of families, get the Pers Genome families, that are families having at least
    tol% of 'nb_strain' members.

    Parameters
    ----------
    fam_by_strain : dict
        {fam_num: {genome1: [members], genome2: [members]}, fam_num2: {genome1: [members]}}
    fam_all_members : dict
        {fam_num: [all members]}
    nb_strains : int
        total number of strains/genomes in dataset
    tol : float
        min percentage of different genomes present in a family
        ex: if tol=50%, and there are 8 genomes. If a family contains 3 genomes, it is not
        persistent. If it contains 7 genomes, it can be persistent (depends on multi and
        mixed parameters)
    multi : bool
        True if multiple genes from the same genome/strain in a family are tolerated. -> a family is considered as multi-persistent if it has members from at least 'tol%' genomes
        False otherwise
    mixed : bool
        True if mixed families are allowed (mixed family = exactly 1 member per genome
        for at least tol% of the genomes, 0 or several members allowed for other (1-tol)% genomes)
    floor : bool
        Use a minimum number of genomes containing a gene to consider the family
        persistent equal to: floor(nb_strains*tol) genomes if True, ceil(nb_strains*tol)
        if False.

    Returns
    -------
    dict
        {fam_num: [list of members]} for persistent families
    """
    logger.info("Generating Persistent genome of a dataset "
                f"containing {nb_strains} genomes")
    pers = {}  # {fam_num: {strain1: [genes from strain1], strain2: [genes from strain2]}}
    fams = {}  # {fam_num: [list of members]}
    if floor:
        min_members = math.floor(tol * nb_strains)
    else:
        min_members = math.ceil(tol * nb_strains)
    for fam_num, family in fam_by_strain.items():
        # If enough strains and multi accepted, or multi not accepted but 1 member per
        # strain, add family to core
        if mixed:
            if mixed_family(family, min_members):
                pers[fam_num] = family
                fams[fam_num] = fam_all_members[fam_num]
        else:
            if len(family) >= min_members and (multi or uniq_members(family)):
                pers[fam_num] = family
                fams[fam_num] = fam_all_members[fam_num]
    # coregenome computed
    if tol == 1 and not multi and not mixed:
        logger.info(f"The core genome contains {len(pers)} families, each one having "
                    f"exactly {int(min_members)} members, from the {nb_strains} different genomes.")
    # multi persistent genome with multigenic families allowed
    elif multi:
        logger.info(f"The persistent genome contains {len(pers)} families with members present "
                    f"in at least {min_members} different genomes ({tol*100}% of the total number of "
                    "genomes).")
    # mixed persistent genome, tol% families with exactly 1 member from each genome,
    # multigenic families allowed for the '1-tol'% remaining families
    elif mixed:
        logger.info(f"The persistent genome contains {len(pers)} families, "
                    f"each one having exactly 1 member from at least {tol*100}% of the genomes ({min_members} "
                    f"genomes). In the remaining {round((1-tol)*100,3)}% genomes, there can be 0, 1 or "
                    "several members.")
    # Strict persistent genome. tol% families with exactly one member in each genome
    else:
        logger.info(f"The persistent genome contains {len(pers)} families, each one having "
                    f"exactly 1 member from at least {tol*100}% of the {nb_strains} "
                    f"different genomes (that is {min_members} genomes). The other genomes are absent from "
                    "the family.")
    return fams


def mixed_family(family, thres):
    """
    1 family = several genomes (genome=strain), each containing x members
    Returns True if at least 'thres' genomes of the family have exactly 1 member.

    Parameters
    ----------
    family : dict
        {strain1: [members in strain1]}
    thres : float
         minimum number of genomes which must have exactly 1 member

    Returns
    -------
    bool
    """
    nb_uniq = 0
    for genome, members in family.items():
        if nb_uniq < thres:
            if len(members) == 0:
                logger.warning(f"Problem, no members for {genome}!")
            if len(members) == 1:
                nb_uniq += 1
        else:
            return True
    return nb_uniq >= thres


def uniq_members(family, num=1):
    """
    Returns True if, in the family, each genome has no more than 'num' member(s),
    False otherwise (multigenic family)

    Parameters
    ----------
    family : dict
        {strain1: [members in strain1], strain2: [members in strain2]}
    num : int
        max number of members allowed in each genome to return True

    Returns
    -------
    bool
    """
    for genome, members in family.items():
        if len(members) == 0:
            logger.warning(f"Problem, no members for {genome}!")
        if len(members) > num:
            return False
    return True


def write_persistent(fams, outfile):
    """
    Write persistent families into output file

    Parameters
    ----------
    fams : dict
        {num_fam: [members]}
    outfile : str
        output file to write all families
    """
    with open(outfile, "w") as outf:
        # Order by family number
        for num_fam in sorted(fams, key=lambda x: int(x)):
            outf.write(str(num_fam))  # Write family num
            fam = fams[num_fam]  # Get list of members of the family
            for mem in sorted(fam, key=utils.sort_proteins):
                outf.write(" " + mem)
            outf.write("\n")



#!/usr/bin/env python3
# coding: utf-8

"""
Unit tests for fastme_func submodule of tree_module
"""

import os
import logging
import pytest
import shutil

import PanACoTA.tree_module.fastme_func as fme
from PanACoTA import utils
import test.test_unit.utilities_for_tests as tutil
from . import utilities as tree_util


# Define common variables
ALPATH = os.path.join("test", "data", "align")
ALIGNMENT = os.path.join(ALPATH, "exp_files", "exp_pers4genomes.grp.aln")
TREEPATH = os.path.join("test", "data", "tree")
EXPPATH = os.path.join(TREEPATH, "exp_files")
GENEPATH = os.path.join(TREEPATH, "generated_by_unit-tests")
LOGFILE_BASE = "log_test_fastme"
LOGFILES = [LOGFILE_BASE + ext for ext in [".log", ".log.debug", ".log.details", ".log.err"]]


@pytest.fixture(autouse=True)
def setup_teardown_module():
    """
    Remove log files at the end of this test module

    Before each test:
    - init logger
    - create directory to put generated files

    After:
    - remove all log files
    - remove directory with generated results
    """
    utils.init_logger(LOGFILE_BASE, 0, 'test_fastme', verbose=1)
    os.mkdir(GENEPATH)
    print("setup")

    yield
    for f in LOGFILES:
        if os.path.exists(f):
            os.remove(f)
    shutil.rmtree(GENEPATH)
    print("teardown")


def test_convert_phylip(caplog):
    """
    Test that when giving a valid fasta alignment file, it converts it to Stockholm format,
    as expected.
    """
    caplog.set_level(logging.DEBUG)
    outfile = os.path.join(GENEPATH, "test_2phylip")
    fme.convert2phylip(ALIGNMENT, outfile)
    exp_stk = os.path.join(EXPPATH, "exp_align_phylip.ph")
    assert os.path.isfile(outfile)
    tutil.compare_order_content(outfile, exp_stk)
    assert "Converting fasta alignment to PHYLIP-relaxed format" in caplog.text


def test_convert_exists(caplog):
    """
    Test that when asking to convert a file in phylip format, but output file already exists,
    it does not convert again, and writes warning message saying that current file will be used.
    """
    caplog.set_level(logging.DEBUG)
    exp_stk = os.path.join(EXPPATH, "exp_align_phylip.ph")
    fme.convert2phylip(ALIGNMENT, exp_stk)
    assert 'Phylip alignment file already existing.' in caplog.text
    assert ("The Phylip alignment file test/data/tree/exp_files/exp_align_phylip.ph "
            "already exists. The program will use it instead of re-converting "
            "test/data/align/exp_files/exp_pers4genomes.grp.aln") in caplog.text


def test_run_fme_default(caplog):
    """
    Test that when running fastme without bootstrap, and with default model, it returns a file
    in the expected format (all branches have lengths, no bootstrap value).
    """
    caplog.set_level(logging.DEBUG)
    source_align = os.path.join(EXPPATH, "exp_align_phylip.ph")
    boot = None
    write_boot = False
    threads = 1
    model = None
    quiet = False
    outdir = GENEPATH
    fme.run_fastme(source_align, boot, write_boot, threads, model, outdir, quiet)

    assert "Running FastME..." in caplog.text
    treefile = os.path.join(GENEPATH, "exp_align_phylip.ph.fastme_tree.nwk")
    assert os.path.isfile(treefile)
    assert ("fastme -i test/data/tree/exp_files/exp_align_phylip.ph -dT -nB -s -T 1 "
            " -o test/data/tree/generated_by_unit-tests/exp_align_phylip.ph.fastme_tree.nwk "
            "-I test/data/tree/generated_by_unit-tests/"
            "exp_align_phylip.ph.fastme.log") in caplog.text
    assert tree_util.is_tree_lengths(treefile)
    assert not tree_util.is_tree_bootstrap(treefile)
    logs = os.path.join(GENEPATH, "exp_align_phylip.ph.fastme.log")
    assert os.path.isfile(logs)


def test_run_fme_boot_j(caplog):
    """
    Test that when running fastme without bootstrap, and with JC69 model, it returns a file
    in the expected format (all branches have lengths, no bootstrap value).
    """
    caplog.set_level(logging.DEBUG)
    source_align = os.path.join(EXPPATH, "exp_align_phylip.ph")
    boot = 105
    write_boot = False
    threads = 1
    model = "J"
    quiet = False
    fme.run_fastme(source_align, boot, write_boot, threads, model, GENEPATH, quiet)
    treefile = os.path.join(GENEPATH, "exp_align_phylip.ph.fastme_tree.nwk")
    assert os.path.isfile(treefile)
    assert "Running FastME..." in caplog.text
    assert ("fastme -i test/data/tree/exp_files/exp_align_phylip.ph -dJ -nB -s -T 1 "
            "-b 105 -o test/data/tree/generated_by_unit-tests/exp_align_phylip.ph.fastme_tree.nwk "
            "-I test/data/tree/generated_by_unit-tests/"
            "exp_align_phylip.ph.fastme.log") in caplog.text
    assert not tree_util.is_tree_lengths(treefile)
    assert tree_util.is_tree_bootstrap(treefile)
    os.remove(os.path.join(EXPPATH, "exp_align_phylip.ph_fastme_boot.txt"))
    logs = os.path.join(GENEPATH, "exp_align_phylip.ph.fastme.log")
    assert os.path.isfile(logs)


def test_run_fme_boot_write_f84(caplog):
    """
    Test that when running fastme with bootstrap, and with F84 model, it returns a file
    in the expected format (all branches have lengths, + bootstrap value).
    + write bootstrap trees
    """
    caplog.set_level(logging.DEBUG)
    source_align = os.path.join(EXPPATH, "exp_align_phylip.ph")
    boot = 105
    write_boot = True
    threads = 1
    model = "4"
    quiet = False
    fme.run_fastme(source_align, boot, write_boot, threads, model, GENEPATH, quiet)

    assert "Running FastME..." in caplog.text
    assert ("fastme -i test/data/tree/exp_files/exp_align_phylip.ph -d4 -nB -s -T 1 "
            "-b 105 "
            "-o test/data/tree/generated_by_unit-tests/exp_align_phylip.ph.fastme_tree.nwk "
            "-I test/data/tree/generated_by_unit-tests/exp_align_phylip.ph.fastme.log "
            "-B test/data/tree/generated_by_unit-tests/"
            "exp_align_phylip.ph.fastme_bootstraps.nwk") in caplog.text
    treefile = os.path.join(GENEPATH, "exp_align_phylip.ph.fastme_tree.nwk")
    assert not tree_util.is_tree_lengths(treefile)
    assert tree_util.is_tree_bootstrap(treefile)
    logs = os.path.join(GENEPATH, "exp_align_phylip.ph.fastme.log")
    assert os.path.isfile(logs)
    bootfile = os.path.join(GENEPATH, "exp_align_phylip.ph.fastme_bootstraps.nwk")
    assert os.path.isfile(bootfile)


def test_run_fme_notreename_rysym(caplog):
    """
    Test that when running fastme without bootstrap, and with RY-symetric model, it returns a file
    in the expected format (all branches have lengths, no bootstrap value).
    """
    caplog.set_level(logging.DEBUG)
    source_align = os.path.join(EXPPATH, "exp_align_phylip.ph")
    boot = 105
    write_boot = True
    threads = 1
    model = "Y"
    quiet = True
    fme.run_fastme(source_align, boot, write_boot, threads, model, GENEPATH, quiet)
    assert "Running FastME..." in caplog.text
    assert ("fastme -i test/data/tree/exp_files/exp_align_phylip.ph -dY -nB -s -T 1 "
            "-b 105 "
            "-o test/data/tree/generated_by_unit-tests/exp_align_phylip.ph.fastme_tree.nwk "
            "-I test/data/tree/generated_by_unit-tests/exp_align_phylip.ph.fastme.log "
            "-B test/data/tree/generated_by_unit-tests/"
            "exp_align_phylip.ph.fastme_bootstraps.nwk") in caplog.text
    treefile = os.path.join(GENEPATH, "exp_align_phylip.ph.fastme_tree.nwk")
    assert not tree_util.is_tree_lengths(treefile)
    assert tree_util.is_tree_bootstrap(treefile)
    logs = os.path.join(GENEPATH, "exp_align_phylip.ph.fastme.log")
    assert os.path.isfile(logs)
    bootfile = os.path.join(GENEPATH, "exp_align_phylip.ph.fastme_bootstraps.nwk")
    assert os.path.isfile(treefile)


def test_run_tree(caplog):
    """
    Test generating tree from fasta alignment with fastme (conversion)
    """
    caplog.set_level(logging.DEBUG)
    boot = None
    quiet = False
    threads = 1
    model = 'T'
    write_boot = False
    fme.run_tree(ALIGNMENT, boot, GENEPATH, quiet, threads, model=model, wb=write_boot)
    assert "Converting fasta alignment to PHYLIP-relaxed format" in caplog.text
    assert "Running FastME..." in caplog.text
    assert ("fastme "
            "-i test/data/tree/generated_by_unit-tests/exp_pers4genomes.grp.aln.phylip "
            "-dT -nB -s -T 1  "
            "-o test/data/tree/generated_by_unit-tests/"
            "exp_pers4genomes.grp.aln.phylip.fastme_tree.nwk "
            "-I test/data/tree/generated_by_unit-tests/"
            "exp_pers4genomes.grp.aln.phylip.fastme.log") in caplog.text
    treefile = os.path.join(GENEPATH, "exp_pers4genomes.grp.aln.phylip.fastme_tree.nwk")
    assert tree_util.is_tree_lengths(treefile)
    assert not tree_util.is_tree_bootstrap(treefile)
    logs = os.path.join(GENEPATH, "exp_pers4genomes.grp.aln.phylip.fastme.log")
    assert os.path.isfile(logs)
    phylip = os.path.join(GENEPATH, "exp_pers4genomes.grp.aln.phylip")
    assert os.path.isfile(phylip)

    # Redo with phylip alignments already generated
    boot = 110
    treefile = os.path.join(GENEPATH, "test_rerun_tree-fastme.tree")
    quiet = False
    threads = 1
    model = 'T'
    write_boot = True
    fme.run_tree(ALIGNMENT, boot, GENEPATH, quiet, threads, model=model, wb=write_boot)
    assert "Phylip alignment file already existing." in caplog.text
    assert ("The Phylip alignment file "
            "test/data/tree/generated_by_unit-tests/exp_pers4genomes.grp.aln.phylip "
            "already exists. The program will use it instead of re-converting "
            "test/data/align/exp_files/exp_pers4genomes.grp.aln") in caplog.text
    assert "Running FastME..." in caplog.text
    assert ("fastme -i test/data/tree/generated_by_unit-tests/exp_pers4genomes.grp.aln.phylip "
            "-dT -nB -s -T 1 -b 110 "
            "-o test/data/tree/generated_by_unit-tests/"
            "exp_pers4genomes.grp.aln.phylip.fastme_tree.nwk "
            "-I test/data/tree/generated_by_unit-tests/"
            "exp_pers4genomes.grp.aln.phylip.fastme.log "
            "-B test/data/tree/generated_by_unit-tests/"
            "exp_pers4genomes.grp.aln.phylip.fastme_bootstraps.nwk") in caplog.text
    treefile = os.path.join(GENEPATH, "exp_pers4genomes.grp.aln.phylip.fastme_tree.nwk")
    assert not tree_util.is_tree_lengths(treefile)
    assert tree_util.is_tree_bootstrap(treefile)
    logs = os.path.join(GENEPATH, "exp_pers4genomes.grp.aln.phylip.fastme.log")
    assert os.path.isfile(logs)
    phylip = os.path.join(GENEPATH, "exp_pers4genomes.grp.aln.phylip")
    assert os.path.isfile(phylip)
    bootfile = "test/data/tree/generated_by_unit-tests/exp_pers4genomes.grp.aln.phylip.fastme_bootstraps.nwk"
    # bootfile = os.path.join(GENEPATH, "exp_pers4genomes.grp.aln.phylip.fastme_bootstraps.nwk")
    assert os.path.isfile(bootfile)




#!/usr/bin/env python3
# coding: utf-8


# ###############################################################################
# This file is part of PanACOTA.                                                #
#                                                                               #
# Authors: Amandine Perrin                                                      #
# Copyright © 2018-2020 Institut Pasteur (Paris).                               #
# See the COPYRIGHT file for details.                                           #
#                                                                               #
# PanACOTA is a software providing tools for large scale bacterial comparative  #
# genomics. From a set of complete and/or draft genomes, you can:               #
#    -  Do a quality control of your strains, to eliminate poor quality         #
# genomes, which would not give any information for the comparative study       #
#    -  Uniformly annotate all genomes                                          #
#    -  Do a Pan-genome                                                         #
#    -  Do a Core or Persistent genome                                          #
#    -  Align all Core/Persistent families                                      #
#    -  Infer a phylogenetic tree from the Core/Persistent families             #
#                                                                               #
# PanACOTA is free software: you can redistribute it and/or modify it under the #
# terms of the Affero GNU General Public License as published by the Free       #
# Software Foundation, either version 3 of the License, or (at your option)     #
# any later version.                                                            #
#                                                                               #
# PanACOTA is distributed in the hope that it will be useful, but WITHOUT ANY   #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS     #
# FOR A PARTICULAR PURPOSE. See the Affero GNU General Public License           #
# for more details.                                                             #
#                                                                               #
# You should have received a copy of the Affero GNU General Public License      #
# along with PanACOTA (COPYING file).                                           #
# If not, see <https://www.gnu.org/licenses/>.                                  #
# ###############################################################################


import sys
from textwrap import dedent

from PanACoTA import __version__ as version

from PanACoTA.subcommands import all_modules
from PanACoTA.subcommands import prepare
from PanACoTA.subcommands import annotate
from PanACoTA.subcommands import pangenome
from PanACoTA.subcommands import corepers
from PanACoTA.subcommands import align
from PanACoTA.subcommands import tree

def main():
    """
    Start program according to arguments given by user.
    """
    action, args = parse_arguments(sys.argv[1:])
    action(args)


def parse_arguments(argv):
    """
    Extract command-line arguments for different actions.
    """
    import argparse

    header = '''



 ___                 _____  ___         _____  _____
(  _`\              (  _  )(  _`\      (_   _)(  _  )
| |_) )  _ _   ___  | (_) || ( (_)   _   | |  | (_) |
| ,__/'/'_` )/' _ `\|  _  || |  _  /'_`\ | |  |  _  |
| |   ( (_| || ( ) || | | || (_( )( (_) )| |  | | | |
(_)   `\__,_)(_) (_)(_) (_)(____/'`\___/'(_)  (_) (_)


       Large scale comparative genomics tools

     -------------------------------------------
     '''

    footer = "For more details, see PanACoTA documentation."

    # Create main parser
    # TITLE with ascii art PanACoTA
    # footer for doc
    parser = argparse.ArgumentParser(epilog=footer,
                                     formatter_class=argparse.RawDescriptionHelpFormatter,
                                     description=dedent(header))


    parser.add_argument('-V', '--version', action='version',
                        version='PanACoTA - v. ' + str(version),
                        help="Print the version number and exit")

    # Create subparsers, for all submodules
    subparsers = parser.add_subparsers(dest='subparser_called')
    # dest: to be able to get the subparser called with args.subparser_called
    actions = {}  # to add the action to do according to the subparser called
    checks = {}  # to add the function to call to check the subparser arguments

    # Running all modules at once. Start with ASCII art title, + small description of subcommand
    parser_all = subparsers.add_parser('all',
                                        formatter_class=argparse.RawDescriptionHelpFormatter,
                                        description=(dedent(header) +
                                        "\n=> Run all PanACoTA modules"),
                                        epilog=footer,
                                        help="Run all PanACoTA modules",
                                        add_help=False)
    all_modules.build_parser(parser_all)
    actions["all"] = all_modules.main_from_parse
    checks["all"] = all_modules.check_args
    # checks["all_modules"] = all_modules.check_args

    # Preparation part. Start with ASCII art title, + small description of subcommand
    parser_prepare = subparsers.add_parser('prepare',
                                            formatter_class=argparse.RawDescriptionHelpFormatter,
                                            description=(dedent(header) +
                                            "\n=> Prepare draft dataset"),
                                            epilog=footer,
                                            help="Prepare draft dataset",
                                            add_help=False)
    prepare.build_parser(parser_prepare)
    actions["prepare"] = prepare.main_from_parse
    checks["prepare"] = prepare.check_args

    # QC and annotation part. Start with ASCII art title, + small description of subcommand
    parser_annotate = subparsers.add_parser('annotate',
                                            formatter_class=argparse.RawDescriptionHelpFormatter,
                                            description=(dedent(header) +
                                            "\n=> Quality control and annotation of genomes"),
                                            epilog=footer,
                                            help="Quality control and annotation of genomes",
                                            add_help=False)
    annotate.build_parser(parser_annotate)
    actions["annotate"] = annotate.main_from_parse
    checks["annotate"] = annotate.check_args

    # Pan genome part. Start with ASCII art title, + small description of subcommand
    parser_pan = subparsers.add_parser('pangenome',
                                       formatter_class=argparse.RawDescriptionHelpFormatter,
                                       description=(dedent(header) +
                                       "\n=> Generate a pan-genome of your dataset"),
                                       epilog=footer, help="Generate a pan-genome of your dataset",
                                       add_help=False)
    pangenome.build_parser(parser_pan)
    actions["pangenome"] = pangenome.main_from_parse

    # Persistent genome part. Start with ASCII art title, + small description of subcommand
    parser_corepers = subparsers.add_parser('corepers',
                                            formatter_class=argparse.RawDescriptionHelpFormatter,
                                            description=(dedent(header) +
                                            "\n=> Compute a Core or Persistent genome of your "
                                            "dataset"),
                                            epilog=footer,
                                            help="Compute a Core or Persistent genome of your "
                                                 "dataset",
                                            add_help=False)
    corepers.build_parser(parser_corepers)
    actions["corepers"] = corepers.main_from_parse
    checks["corepers"] = corepers.check_args

    # Alignment part. Start with ASCII art title, + small description of subcommand
    parser_align = subparsers.add_parser('align',
                                         formatter_class=argparse.RawDescriptionHelpFormatter,
                                         description=(dedent(header) +
                                         "\n=> Align Core/Persistent families"),
                                         epilog=footer,
                                         help="Align Core/Persistent families",
                                         add_help=False)
    align.build_parser(parser_align)
    actions["align"] = align.main_from_parse

    # tree part. Start with ASCII art title, + small description of subcommand
    parser_tree = subparsers.add_parser('tree',
                                        formatter_class=argparse.RawDescriptionHelpFormatter,
                                        description=(dedent(header) +
                                        "\n=> Infer phylogenetic tree based on "
                                        "core/persistent genome"),
                                        epilog=footer,
                                        help="Infer phylogenetic tree based on "
                                             "core/persistent genome",
                                        add_help=False)
    tree.build_parser(parser_tree)
    actions["tree"] = tree.main_from_parse
    checks["tree"] = tree.check_args

    # Parse arguments and execute corresponding action
    arguments = parser.parse_args(argv)
    arguments.argv = argv
    action_called = arguments.subparser_called
    # If checks are needed, do it (if some arguments are not compatible etc.)
    if action_called in checks:
        checks[action_called](parser, arguments)

    # If subparser called does not exist, error
    if action_called not in actions:
        parser.error("too few arguments. Use '-h' to get help.")
    return actions[action_called], arguments


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# coding: utf-8

# ###############################################################################
# This file is part of PanACOTA.                                                #
#                                                                               #
# Authors: Amandine Perrin                                                      #
# Copyright © 2018-2020 Institut Pasteur (Paris).                               #
# See the COPYRIGHT file for details.                                           #
#                                                                               #
# PanACOTA is a software providing tools for large scale bacterial comparative  #
# genomics. From a set of complete and/or draft genomes, you can:               #
#    -  Do a quality control of your strains, to eliminate poor quality         #
# genomes, which would not give any information for the comparative study       #
#    -  Uniformly annotate all genomes                                          #
#    -  Do a Pan-genome                                                         #
#    -  Do a Core or Persistent genome                                          #
#    -  Align all Core/Persistent families                                      #
#    -  Infer a phylogenetic tree from the Core/Persistent families             #
#                                                                               #
# PanACOTA is free software: you can redistribute it and/or modify it under the #
# terms of the Affero GNU General Public License as published by the Free       #
# Software Foundation, either version 3 of the License, or (at your option)     #
# any later version.                                                            #
#                                                                               #
# PanACOTA is distributed in the hope that it will be useful, but WITHOUT ANY   #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS     #
# FOR A PARTICULAR PURPOSE. See the Affero GNU General Public License           #
# for more details.                                                             #
#                                                                               #
# You should have received a copy of the Affero GNU General Public License      #
# along with PanACOTA (COPYING file).                                           #
# If not, see <https://www.gnu.org/licenses/>.                                  #
# ###############################################################################

"""
Functions to check argparse aguments given by user


@author gem
April 2017
"""
from PanACoTA import utils
import argparse
import configparser
import sys
import os


def gen_name(param):
    """
    Check name given for the genome set:
    - must contain 4 characters, all alphanumeric (calling utils.check_format)
    - if does not fit, give advice on format (genus_genus_species_species)
    """
    if not utils.check_format(param):
        msg = ("The genome name must contain 4 characters. For example, this name can "
               "correspond to the 2 first letters of genus, and 2 first letters of "
               "species, e.g. ESCO for Escherichia Coli.")
        raise argparse.ArgumentTypeError(msg)
    return param


def date_name(param):
    """
    Check format of given date:
    - must contain 4 characters, all alphanumeric (calling utils.check_format)
    - if does not fit, give advice on format (MMYY)
    """
    if not utils.check_format(param):
        msg = ("The date must contain 4 characters. Usually, it contains 4 digits, "
               "corresponding to the month (2 digits) and year (2 digits).")
        raise argparse.ArgumentTypeError(msg)
    return param


def get_date():
    """
    Get current date in MMYY format
    """
    import time
    return time.strftime("%m%y")


def cont_num(param):
    """
    Check number of contigs given
    - must be positive int
    - not more than 10000
    """
    try:
        param = int(param)
    except Exception:
        msg = "argument --nbcont: invalid int value: {}".format(param)
        raise argparse.ArgumentTypeError(msg)
    if param < 0:
        msg = "The maximum number of contigs allowed must be a positive number."
        raise argparse.ArgumentTypeError(msg)
    if param >= 10000:
        msg = "We do not support genomes with more than 9999 contigs."
        raise argparse.ArgumentTypeError(msg)
    return param


def thread_num(param):
    """
    check number of threads given.
    - must be a positive int
    - Cannot be more than number of threads available
    - if '0' given, return number of threads available
    """
    import multiprocessing
    try:
        param = int(param)
    except Exception:
        msg = "argument --threads threads: invalid int value: {}".format(param)
        raise argparse.ArgumentTypeError(msg)
    nb_cpu = multiprocessing.cpu_count()
    if param > nb_cpu:
        msg = ("You have {} threads on your computer, you cannot ask for more: "
               "invalid value: {}").format(nb_cpu, param)
        raise argparse.ArgumentTypeError(msg)
    elif param < 0:
        msg = ("Please provide a positive number of threads (or 0 for all threads): "
               "Invalid value: {}").format(param)
        raise argparse.ArgumentTypeError(msg)
    elif param == 0:
        return nb_cpu
    return param


def positive_int(param):
    """
    Return a positive int for argument --cutn
    """
    try:
        param = int(param)
    except ValueError:
        msg = f"error: argument --cutn: invalid int value: '{param}'"
        raise argparse.ArgumentTypeError(msg)
    if param < 0:
        msg = f"error: argument --cutn must be a positive integer: invalid int value: '{param}'"
        raise argparse.ArgumentTypeError(msg)
    return param


def mash_dist(param):
    """
    Check mash distance given. Must be a float between 0 and 1 included
    """
    try:
        param = float(param)
    except ValueError:
        msg = f"error: mash distance: invalid float value: '{param}'"
        raise argparse.ArgumentTypeError(msg)
    if param < 0 or param > 1:
        msg = f"error: mash distance must be between 0 and 1: invalid value: '{param}'"
        raise argparse.ArgumentTypeError(msg)
    return param


def percentage(param):
    """
    check argument given to parameter '-t tol'
    """
    try:
        param = float(param)
    except Exception:
        msg = "argument -t tol: invalid float value: {}".format(param)
        raise argparse.ArgumentTypeError(msg)
    if param < 0 or param > 1:
        msg = ("The minimum %% of genomes required in a family to be persistent must "
               "be in [0, 1]. Invalid value: {}".format(param))
        raise argparse.ArgumentTypeError(msg)
    return param


def perc_id(param):
    """
    Check argument given to parameter -i percentage_id
    """
    try:
        param = float(param)
    except Exception:
        msg = "argument -i percentage_id: invalid float value: {}".format(param)
        raise argparse.ArgumentTypeError(msg)
    if param < 0 or param > 1:
        msg = ("The minimum %% of identity must be in [0, 1]. Invalid value: {}".format(param))
        raise argparse.ArgumentTypeError(msg)
    return param


class Conf_all_parser(configparser.ConfigParser):
    """
    Read configfile and return arguments found, according to required type
    """
    def __init__(self, conffile, readsec=[]):
        super().__init__()
        if conffile != "" and not os.path.isfile(conffile):
            print(f"Error: config file {conffile} not found.")
            sys.exit(1)
        self.conffile = conffile
        try:
            self.read(conffile)
        except configparser.DuplicateOptionError as err:
            print(err)
            sys.exit(1)
        self.sec_dicts = {}
        for sec in readsec:
            # If section in configfile, put its arguments and values to a dict
            # If not, create empty section, and associate with empty dict
            if sec in dict(self):
                self.sec_dicts[sec] = dict(self[sec])
            else:
                self.sec_dicts[sec] = {}
                self.add_section(sec)

    def get_section_dict(self, section):
        """
        get dictionary of values for 'section' section
        """
        if section in self.sec_dicts:
            return self.sec_dicts[section]
        else:
            print(f"No section {section} in {self.conffile}")
            sys.exit(1)

    def add_default(self, defargs, section):
        """
        Complete 'section' dict with default parameters.
        If key already defined, keep current value.
        """
        for key, val in defargs.items():
            if key not in self.sec_dicts[section]:
                self[section][key] = str(val)
                self.sec_dicts[section][key] = val

    def update(self, args, section):
        """
        Add all arguments from args. If key already exists in self, overwrite it.
        Otherwise, create it.
        """
        self.sec_dicts[section].update(args)
        for key, val in self.sec_dicts[section].items():
            self[section][key] = str(val)

    def set_boolean(self, section, param):
        """
        Change param of section to boolean
        raise error if problem
        """
        try:
            bool_param = self.getboolean(section, param)
            self.sec_dicts[section][param] = bool_param
        except ValueError as err:
            val = self[section][param]
            print(f"ERROR: {param} must be a boolean. Wrong value: {val}.")
            sys.exit(1)

    def set_int(self, section, param):
        """
        Change param of section to int
        raise error if problem
        """
        try:
            int_param = self.getint(section, param)
            self.sec_dicts[section][param] = int_param
        except ValueError as err:
            val = self[section][param]
            print(f"ERROR: {param} must be an int. Wrong value: {val}.")
            sys.exit(1)

    def set_float(self, section, param):
        """
        Change param of section to float
        raise error if problem
        """
        try:
            float_param = self.getfloat(section, param)
            self.sec_dicts[section][param] = float_param
        except ValueError as err:
            val = self[section][param]
            print(f"ERROR: {param} must be a float. Wrong value: {val}.")
            sys.exit(1)



#!/usr/bin/env python3
# coding: utf-8

# ###############################################################################
# This file is part of PanACOTA.                                                #
#                                                                               #
# Authors: Amandine Perrin                                                      #
# Copyright © 2018-2020 Institut Pasteur (Paris).                               #
# See the COPYRIGHT file for details.                                           #
#                                                                               #
# PanACOTA is a software providing tools for large scale bacterial comparative  #
# genomics. From a set of complete and/or draft genomes, you can:               #
#    -  Do a quality control of your strains, to eliminate poor quality         #
# genomes, which would not give any information for the comparative study       #
#    -  Uniformly annotate all genomes                                          #
#    -  Do a Pan-genome                                                         #
#    -  Do a Core or Persistent genome                                          #
#    -  Align all Core/Persistent families                                      #
#    -  Infer a phylogenetic tree from the Core/Persistent families             #
#                                                                               #
# PanACOTA is free software: you can redistribute it and/or modify it under the #
# terms of the Affero GNU General Public License as published by the Free       #
# Software Foundation, either version 3 of the License, or (at your option)     #
# any later version.                                                            #
#                                                                               #
# PanACOTA is distributed in the hope that it will be useful, but WITHOUT ANY   #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS     #
# FOR A PARTICULAR PURPOSE. See the Affero GNU General Public License           #
# for more details.                                                             #
#                                                                               #
# You should have received a copy of the Affero GNU General Public License      #
# along with PanACOTA (COPYING file).                                           #
# If not, see <https://www.gnu.org/licenses/>.                                  #
# ###############################################################################

"""
Functions to deal with prokka or prodigal only, according to user request


@author gem
April 2017
"""

import os
import sys
import shutil
import glob
import logging
import subprocess
import shlex
import multiprocessing
import progressbar
import threading

import PanACoTA.utils as utils

logger = logging.getLogger('annotate.run_annotation_all')


def run_annotation_all(genomes, threads, force, annot_folder, fgn, prodigal_only=False,
                       small=False, quiet=False):
    """
    For each genome in genomes, run prokka (or only prodigal) to annotate the genome.

    Parameters
    ----------
    genomes : dict
        {genome: [gembase_name, path_to_origfile, path_split_gembase, gsize, nbcont, L90]}
    threads : int
        max number of threads that can be used
    force : bool
        if False, do not override prokka/prodigal outdir and result dir if they exist. If\
        True, rerun prokka and override existing results, for all genomes.
    annot_folder : str
        folder where prokka/prodigal results must be written: for each genome,
        a directory <genome_name>-prokkaRes or <genome_name>-prodigalRes> will be created
        in this folder, and all the results
        of prokka/prodigal for the genome will be written inside
    fgn : str
        name (key in genomes dict) of the fist genome, which will be used for prodigal training
    prodigal_only : bool
        True if only prodigal must run, False if prokka must run
    small : bool
        True -> use -p meta option with prodigal. Do not use training
    quiet : bool
        True if nothing must be written to stderr/stdout, False otherwise

    Returns
    -------
    dict
        {genome: boolean} -> with True if prokka/prodigal ran well, False otherwise.
    """

    # Update information according to annotation soft used and write message
    if prodigal_only:
        message = "Annotating all genomes with prodigal"
        run_annot = run_prodigal
        main_logger = logging.getLogger("annotate.prodigal")
    else:
        message = "Annotating all genomes with prokka"
        run_annot = run_prokka
        main_logger = logging.getLogger("annotate.prokka")
    main_logger.info(message)
    # Get total number of genomes to annotate, used to show annotation progress
    nbgen = len(genomes)
    bar = None
    # If user did not ask for quiet, create progressbar
    if not quiet:
        # Create progress bar
        widgets = ['Annotation: ', progressbar.Bar(marker='█', left='', right=''),
                   ' ', progressbar.Counter(), "/{}".format(nbgen), ' (',
                   progressbar.Percentage(), ') - ', progressbar.Timer(), ' - '
                  ]
        bar = progressbar.ProgressBar(widgets=widgets, max_value=nbgen,
                                      term_width=79).start()
    # Get resource availability:
    # - number of threads used by prokka/prodigal (cores_annot)
    # - how many genomes can be annotated at the same time (pool_size)
    # - prodigal does not run with several threads: with prodigal, always cores_annot == 1
    # and pool_size == threads
    gpath_train = ""  # by default, no training genome
    if prodigal_only:
        cores_annot = 1
        pool_size = threads
        # If prodigal, train on the first genome
        # fgn is key of genomes, genomes[fgn] = [_,_,annote_file,_,_,_]
        gtrain = genomes[fgn][2]
        # If problem, gpath_train will be empty, but this will be checked while
        # trying to run prodigal, because we also need to check that genomes are not simply
        # already annotated
        if not small:
            gpath_train = prodigal_train(gtrain, annot_folder)
        else:
            gpath_train = "small option"
    elif threads <= 3:
        # less than 3 threads: run prokka 1 by 1 with all threads
        cores_annot = threads
        pool_size = 1
    else:
        # use multiprocessing
        # if there are more threads than genomes, use as many threads as possible per genome
        if len(genomes) <= threads:
            cores_annot = int(threads / len(genomes))
        # otherwise, use 2 threads per genome (and nb_genome/2 genomes at the same time)
        else:
            cores_annot = 2
        pool_size = int(threads / cores_annot)
    #  Create pool with a given size (=number of tasks to be launched in parallel)
    pool = multiprocessing.Pool(pool_size)
    # Create a Queue to put logs from processes, and handle them after from a single thread
    m = multiprocessing.Manager()
    q = m.Queue()
    # {genome: [gembase_name, path_to_origfile, path_toannotate_file, gsize, nbcont, L90]}
    # arguments: gpath, prok_folder, threads, name, force, nbcont, small(for prodigal), q
    arguments = [(genomes[g][2], annot_folder, cores_annot, genomes[g][0],
                  force, genomes[g][4], gpath_train, q)
                 for g in sorted(genomes)]
    try:
        # Start pool (run 'run_annot' n each set of arguments)
        final = pool.map_async(run_annot, arguments, chunksize=1)
        # Close pool: no more data will be put on this pool
        pool.close()
        # Listen for logs in processes
        lp = threading.Thread(target=utils.logger_thread, args=(q,))
        lp.start()
        if not quiet:
            while True:
                # Final is ready when all pool elements are done
                if final.ready():
                    break
                # If not done, get number of genomes left
                remaining = final._number_left
                # Add this to start progressbar with 0% instead of N/A%
                if remaining == nbgen:
                    bar.update(0.0000001)
                else:
                    # Update progress bar
                    bar.update(nbgen - remaining)
            # End progress bar
            bar.finish()
        pool.join()
        # Put None to tell 'q' that everything is finished. It can stopped and be joined.
        q.put(None)
        # join lp (tell to stop once all log processes are done, which is the case here)
        lp.join()
        final = final.get()
    # # If user stops programm (ctrl+C), end it
    # except KeyboardInterrupt as ki:
    #     print("error")
    #     for worker in pool._pool:
    #         print(worker.is_alive())
    #     pool.join()
    #     print("closed")
    #     pool.terminate()
    #     print("--------------terminate ok----------------")
    #     lp.join()
    #     print("thread stopped")
    #     # run_event.clear()
    #     # lp.terminate()
    #     # print("--------------JOIN--------------")
    #     # pool.terminate()
    #     main_logger.error("Process killed by CTRL+C")
    #     return "coucou"
    # If an error occurs, terminate pool, write error and exit
    except Exception as excp:  # pragma: no cover
        pool.terminate()
        main_logger.error(excp)
        sys.exit(1)
    final = {genome: res for genome, res in zip(sorted(genomes), final)}
    return final


def prodigal_train(gpath, annot_folder):
    """
    Use prodigal training mode.
    First, train prodigal on the first genome ('gpath'), and write it to 'genome'.trn,
    file which will be used for the annotation of all next sequence
    Parameters
    ----------
    gpath : str
        path to genome to train on
    annot_folder : str
        path to folder where the log files and train file will be saved

    Returns
    -------
    str
        path and name of train file (will be used to annotate all next genomes)
        If problem, returns empty string
    """
    logger.info(f"Prodigal will train using {gpath}")
    gname = os.path.basename(gpath)             # path/to/original/genome.fasta -> genome.fasta
    gpath_train = os.path.join(annot_folder, gname + ".trn") # path/to/prodiRes/genome.fasta.trn
    if os.path.isfile(gpath_train):
        logger.info(f"A training file already exists ({gpath_train}). "
                     "It will be used to annotate all genomes.")
        return gpath_train
    prodigal_logfile = gpath_train + "-prodigal-train.log"  # path/to/genome-prodigal-train.log
    prodigal_logfile_err = gpath_train + "-prodigal-train.log.err"
    cmd = (f"prodigal -i {gpath} -t {gpath_train}")
    error = (f"Error while trying to train prodigal on {gname}. See {prodigal_logfile_err}.")
    logger.log(utils.detail_lvl(), "prodigal command: " + cmd)
    prodigalf = open(prodigal_logfile, "w")
    prodigalferr = open(prodigal_logfile_err, "w")
    ret = utils.run_cmd(cmd, error, eof=False, stderr=prodigalferr, stdout=prodigalf,
                        logger=logger)
    prodigalf.close()
    prodigalferr.close()
    if ret.returncode == 0:
        logger.log(utils.detail_lvl(), f"End training on {gpath}")
        return gpath_train
    else:
        return ""


def run_prokka(arguments):
    """
    Run prokka for the given genome.

    Parameters
    ----------
    arguments : tuple
        (gpath, prok_folder, cores_annot, name, force, nbcont, small, q) with:

        * gpath: path and filename of genome to annotate
        * prok_folder: path to folder where all prokka folders for all genomes are saved
        * cores_annot: how many cores can use prokka
        * name: output name of annotated genome
        * force: True if force run (override existing files), False otherwise
        * nbcont: number of contigs in the input genome, to check prokka results
        * small: used for prodigal, if sequences to annotate are small. Not used here
        * q : queue where logs are put

    Returns
    -------
    boolean
        True if eveything went well (all needed output files present,
        corresponding numbers of proteins, genes etc.). False otherwise.
    """
    gpath, prok_folder, threads, name, force, nbcont, _, q = arguments
    # Set logger for this process
    qh = logging.handlers.QueueHandler(q)
    root = logging.getLogger()
    root.setLevel(logging.DEBUG)
    root.handlers = []
    logging.addLevelName(utils.detail_lvl(), "DETAIL")
    root.addHandler(qh)
    logger = logging.getLogger('annotate.run_prokka')
    logger.log(utils.detail_lvl(), f"Start annotating {name} from {gpath} with Prokka")

    # Define prokka directory and logfile, and check their existence
    prok_dir = os.path.join(prok_folder, os.path.basename(gpath) + "-prokkaRes")
    fnull = open(os.devnull, 'w')
    prok_logfile = os.path.join(prok_folder, os.path.basename(gpath) + "-prokka.log")
    # import sys
    # sys.exit(1)
    # If result dir already exists, check if we can use it or next step or not
    if os.path.isdir(prok_dir) and not force:
        logger.warning(f"Prokka results folder {prok_dir} already exists.")
        ok = check_prokka(prok_dir, prok_logfile, name, gpath, nbcont, logger)
        # If everything ok in the result dir, do not rerun prokka,
        # use those results for next step (formatting)
        if ok:
            logger.log(utils.detail_lvl(), "Prokka did not run again, "
                       "formatting step used already generated results of "
                       f"Prokka in {prok_dir}. If you want to re-run prokka, first "
                       "remove this result folder, or use '-F' or '--force' "
                       "option if you want to rerun prokka for all genomes.")
            logger.log(utils.detail_lvl(), f"End annotating {name} {gpath}")
        # If missing files, or other problems in result dir, error message,
        # ask user to force or remove this folder.
        else:
            logger.warning("Problems in the files contained in your already existing output dir "
                           "({}). Please check it, or remove it to "
                           "re-annotate.".format(prok_dir))
        # If everything was ok -> everything is ready for next step -> return True
        # If something is wrong -> cannot use those results, genome won't be annotated
        # -> return False
        return ok
    # If result dir exists but user wants to force, remove this result dir
    elif os.path.isdir(prok_dir) and force:
        shutil.rmtree(prok_dir)
        logger.warning("Prokka results folder already exists, but removed because --force option "
                       "used")
    # Now that we checked and solved those cases:
    #     - outdir exists (problems or not, we returned appropriate boolean)
    #     - if outdir exists exists but force, remove this outdir.
    # So, outdir does not exist -> run prokka
    cmd = (f"prokka --outdir {prok_dir} --cpus {threads} "
           f"--prefix {name} --centre prokka {gpath}")
    error = (f"Error while trying to run prokka on {name} from {gpath}")
    logger.log(utils.detail_lvl(), "Prokka command: " + cmd)
    prokf = open(prok_logfile, "w")
    ret = utils.run_cmd(cmd, error, eof=False, stderr=prokf, logger=logger)
    prokf.close()
    if ret.returncode != 0:
        return False
    ok = check_prokka(prok_dir, prok_logfile, name, gpath, nbcont, logger)
    logger.log(utils.detail_lvl(), f"End annotating {name} from {gpath}.")
    return ok


def run_prodigal(arguments):
    """
    Run prodigal for the given genome.

    Parameters
    ----------
    arguments : tuple
        (gpath, prodigal_folder, cores_annot, name, force, nbcont, q) with:

        * gpath: path and filename of genome to annotate
        * prodigal_folder: path to folder where all prodigal folders for all genomes are saved
        * cores_annot: how many cores can use prodigal
        * name: output name of annotated genome
        * force: True if force run (override existing files), False otherwise
        * nbcont: number of contigs in the input genome, to check prodigal results
        * small: ifcontigs are too small (<20000bp), use -p meta option
        * q : queue where logs are put

    Returns
    -------
    boolean
        True if eveything went well (all needed output files present,
        corresponding numbers of proteins, genes etc.). False otherwise.
    """
    gpath, prodigal_folder, threads, name, force, nbcont, gpath_train, q = arguments
    # Set logger for this process, which will be given to all subprocess
    qh = logging.handlers.QueueHandler(q)
    root = logging.getLogger()
    root.setLevel(logging.DEBUG)
    root.handlers = []
    logging.addLevelName(utils.detail_lvl(), "DETAIL")
    root.addHandler(qh)
    logger = logging.getLogger('annotate.run_prodigal')
    # Define prodigal directory and logfile, and check their existence
    # By default, prodigal is in tmp_folder -> resdir/tmp_files/genome-prodigalRes
    g_ori_name = os.path.basename(gpath)
    prodigal_dir = os.path.join(prodigal_folder, g_ori_name + "-prodigalRes")
    prodigal_logfile = os.path.join(prodigal_folder, g_ori_name + "-prodigal.log")
    prodigal_logfile_err = os.path.join(prodigal_folder, g_ori_name + "-prodigal.log.err")

    # If result dir exists but user wants to force, remove this result dir
    if os.path.isdir(prodigal_dir) and force:
        shutil.rmtree(prodigal_dir)
        logger.warning("Prodigal results folder already exists, but is removed because "
                       "--force option was used.")

    # Training file can be "small option", meaning that we did not use the training mode.
    # If not "small option", we used the training mode. If training file does not exist 
    # and prodigal result directory neither, return False
    # We cannot annotate using nothing.
    # Happens if there was a problem while training
    if (gpath_train != "small option" and not os.path.isfile(gpath_train) 
        and not os.path.isdir(prodigal_dir)):
        return False

    logger.log(utils.detail_lvl(), f"Start annotating {name} (from {gpath} sequence) "
                                     "with Prodigal")
    # If prodigal results dir already exists (meaning user did not want to force,
    # otherwise it would have been deleted just before),
    # can we use it for next step ? -> check content.
    if os.path.isdir(prodigal_dir):
        logger.warning(f"Prodigal results folder {prodigal_dir} already exists.")
        ok = check_prodigal(gpath, name, prodigal_dir, logger)
        # If everything ok in the result dir, do not rerun prodigal,
        # use those results for next step (formatting)
        if ok:
            logger.log(utils.detail_lvl(), "Prodigal did not run again. "
                                           "Formatting step will use already generated results of "
                                           "Prodigal in {}. If you want to re-run Prodigal, first "
                                           "remove this result folder, or use '-F' or '--force' "
                                           "option.".format(prodigal_dir))

            logger.log(utils.detail_lvl(), f"End annotating {name} (from {gpath})")
        # If missing files, or other problems in result dir, error message,
        # ask user to force or remove this folder.
        else:
            logger.warning("Problems in the files contained in your already existing output dir "
                           f"({prodigal_dir}). Please check it, or remove it to "
                           "re-annotate.")
        # If everything was ok -> everything is ready for next step -> return True
        # If something is wrong -> cannot use those results, genome won't be annotated
        # -> return False
        return ok
    else:
        # We are sure prodigal result dir does not exist yet, because either:
        #     - never existed
        #     - removed because user asked to force
        #     - exists but left function, so does not go until this line
        #        -> either if files inside are ok or not
        # So make prodigal_dir (not automatically created by prodigal)
        os.makedirs(prodigal_dir)

    # Prodigal_directory is empty and ready to get prodigal results
    basic_outname = os.path.join(prodigal_dir, name)
    # Define cmd, stderr and stdout files, and error to write if problem.
    error = (f"Error while trying to run prodigal. See {prodigal_logfile_err}.")
    prodigalf = open(prodigal_logfile, "w")
    prodigalferr = open(prodigal_logfile_err, "w")
    if gpath_train == "small option":
        training = "-p meta"
    else:
        training = f"-t {gpath_train}"
    cmd = (f"prodigal -i {gpath} -d {basic_outname + '.ffn'} -a {basic_outname + '.faa'} "
           f"-f gff -o {basic_outname + '.gff'} {training} -q")
    logger.log(utils.detail_lvl(), "Prodigal command: " + cmd)

    ret = utils.run_cmd(cmd, error, eof=False, stderr=prodigalferr, stdout=prodigalf,
                        logger=logger)
    prodigalf.close()
    prodigalferr.close()
    if ret.returncode == 0:
        logger.log(utils.detail_lvl(), f"End annotating {name} (from {gpath})")
        return True
    else:
        return False


def check_prokka(outdir, logf, name, gpath, nbcont, logger):
    """
    Prokka writes everything to stderr, and always returns a non-zero return code. So, we
    check if it ran well by checking the content of output directory.
    This function is also used when prokka files already exist (prokka was run previously), to
    check if everything is ok before going to next step.

    Parameters
    ----------
    outdir : str
        output directory, where all files are written by prokka
    logf : str
        prokka/prodigal logfile, containing stderr of prokka
    name : str
        genome name in gembase format
    gpath : str
        path to fasta file given as input for prokka
    nbcont : int
        number of contigs in fasta file given to prokka
    logger : logging.Logger
        logger object to get logs

    Returns
    -------
    bool
        True if everything went well, False otherwise
    """
    missing_file = False
    problem = False
    if not os.path.isdir(outdir):
        logger.error(("Previous annotation could not run properly. "
                      "Look at {} for more information.").format(logf))
        missing_file = True
    else:
        oriname = os.path.basename(gpath)
        fnafile = glob.glob(os.path.join(outdir, "*.fna"))
        tblfile = glob.glob(os.path.join(outdir, "*.tbl"))
        faafile = glob.glob(os.path.join(outdir, "*.faa"))
        ffnfile = glob.glob(os.path.join(outdir, "*.ffn"))
        gfffile = glob.glob(os.path.join(outdir, "*.gff"))
        if len(fnafile) == 0:
            logger.error("{} {}: no .fna file".format(name, oriname))
            missing_file = True
        elif len(fnafile) > 1:
            logger.error("{} {}: several .fna files".format(name, oriname))
            missing_file = True
        if len(tblfile) == 0:
            logger.error("{} {}: no .tbl file".format(name, oriname))
            missing_file = True
        elif len(tblfile) > 1:
            logger.error("{} {}: several .tbl files".format(name, oriname))
            missing_file = True
        if len(faafile) == 0:
            logger.error("{} {}: no .faa file".format(name, oriname))
            missing_file = True
        elif len(faafile) > 1:
            logger.error("{} {}: several .faa files".format(name, oriname))
            missing_file = True
        if len(ffnfile) == 0:
            logger.error("{} {}: no .ffn file".format(name, oriname))
            missing_file = True
        elif len(ffnfile) > 1:
            logger.error("{} {}: several .ffn files".format(name, oriname))
            missing_file = True
        if len(gfffile) == 0:
            logger.error("{} {}: no .gff file".format(name, oriname))
            missing_file = True
        elif len(gfffile) > 1:
            logger.error("{} {}: several .gff files".format(name, oriname))
            missing_file = True
        if not missing_file:
            tblfile = tblfile[0]
            faafile = faafile[0]
            ffnfile = ffnfile[0]
            fnbcont, tnb_cds, nb_gene = count_tbl(tblfile)
            faaprot = count_headers(faafile)
            ffngene = count_headers(ffnfile)
            if nbcont != fnbcont:
                logger.error(("{} {}: no matching number of contigs; "
                              "nbcontig={}; in tbl ={}").format(name, oriname, nbcont, fnbcont))
                problem = True
            if tnb_cds != faaprot:
                logger.error(("{} {}: no matching number of proteins between tbl and faa; "
                              "faa={}; in tbl ={}").format(name, oriname, faaprot, tnb_cds))
                problem = True
    return not problem and not missing_file


def check_prodigal(gpath, name, prodigal_dir, logger):
    """
    When prodigal result folder already exists, check that the ouput files exist.
    We cannot check all content, but check that they are present.

    Parameters
    ----------
    gpath : str
        path to fasta file given as input for prokka
    name : str
        genome name in gembase format
    prodigal_dir : str
        output directory, where all files are written by prodigal
    logger : logging.Logger
        logger object to get logs

    Returns
    -------
    bool
        True if everything went well, False otherwise
    """
    oriname = os.path.basename(gpath)
    faafile = glob.glob(os.path.join(prodigal_dir, "*.faa"))
    ffnfile = glob.glob(os.path.join(prodigal_dir, "*.ffn"))
    gfffile = glob.glob(os.path.join(prodigal_dir, "*.gff"))
    missing_file = False

    if len(faafile) != 1:
        logger.error("{} {}: no or several .faa file(s)".format(name, oriname))
        missing_file = True
    if len(ffnfile) !=  1:
        logger.error("{} {}: no or several .ffn file(s)".format(name, oriname))
        missing_file = True
    if len(gfffile) != 1:
        logger.error("{} {}: no or several .gff file(s)".format(name, oriname))
        missing_file = True

    # If we have all result files, check they are not empty
    if not missing_file:
        if (os.path.getsize(faafile[0]) == 0 or os.path.getsize(ffnfile[0]) == 0
            or os.path.getsize(gfffile[0]) == 0):
            logger.error("Genome {} (from {}): At least one of your Prodigal result file "
                         "is empty.".format(name, oriname))
            return False
    return not missing_file


def count_tbl(tblfile):
    """
    Count the different features found in the tbl file:

    - number of contigs
    - number of proteins (CDS)
    - number of genes (locus_tag)
    - number of CRISPR arrays (repeat_region) -> ignore crisprs

    Parameters
    ----------
    tblfile : str
        tbl file generated by prokka

    Returns
    -------
    (nbcont, nb_cds, nb_gene, nb_crispr)
        information on features found in the tbl file.
    """
    nbcont = 0
    nb_cds = 0
    nb_gene = 0
    # nb_crispr = 0
    with open(tblfile) as tblf:
        for line in tblf:
            if line.startswith(">"):
                nbcont += 1
            if "CDS" in line:
                nb_cds += 1
            if "locus_tag" in line:
                nb_gene += 1
            # if "repeat_region" in line or (len(line.split()) == 3 and "CRISPR" in line):
            #     nb_crispr += 1
    return nbcont, nb_cds, nb_gene  #, nb_crispr


def count_headers(seqfile):
    """
    Count how many sequences there are in the given file

    Parameters
    ----------
    seqfile : str
        file containing a sequence in multi-fasta format

    Returns
    -------
    int
        number of contigs in the given multi-fasta file
    """
    nbseq = 0
    with open(seqfile) as faaf:
        for line in faaf:
            if line.startswith(">"):
                nbseq += 1
    return nbseq



#!/usr/bin/env python3
# coding: utf-8

# ###############################################################################
# This file is part of PanACOTA.                                                #
#                                                                               #
# Authors: Amandine Perrin                                                      #
# Copyright © 2018-2020 Institut Pasteur (Paris).                               #
# See the COPYRIGHT file for details.                                           #
#                                                                               #
# PanACOTA is a software providing tools for large scale bacterial comparative  #
# genomics. From a set of complete and/or draft genomes, you can:               #
#    -  Do a quality control of your strains, to eliminate poor quality         #
# genomes, which would not give any information for the comparative study       #
#    -  Uniformly annotate all genomes                                          #
#    -  Do a Pan-genome                                                         #
#    -  Do a Core or Persistent genome                                          #
#    -  Align all Core/Persistent families                                      #
#    -  Infer a phylogenetic tree from the Core/Persistent families             #
#                                                                               #
# PanACOTA is free software: you can redistribute it and/or modify it under the #
# terms of the Affero GNU General Public License as published by the Free       #
# Software Foundation, either version 3 of the License, or (at your option)     #
# any later version.                                                            #
#                                                                               #
# PanACOTA is distributed in the hope that it will be useful, but WITHOUT ANY   #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS     #
# FOR A PARTICULAR PURPOSE. See the Affero GNU General Public License           #
# for more details.                                                             #
#                                                                               #
# You should have received a copy of the Affero GNU General Public License      #
# along with PanACOTA (COPYING file).                                           #
# If not, see <https://www.gnu.org/licenses/>.                                  #
# ###############################################################################

"""
pangenome is a subcommand of PanACoTA


@author gem
May 2017
"""

import sys
import os


def main_from_parse(args):
    """
    Call main function from the arguments given by parser

    Parameters
    ----------
    args : argparse.Namespace
        result of argparse parsing of all arguments in command line
    """
    cmd = "PanACoTA " + ' '.join(args.argv)
    main(cmd, args.lstinfo_file, args.dataset_name, args.dbpath, args.min_id, args.outdir,
         args.clust_mode, args.spedir, args.threads, args.outfile, args.verbose,
         args.quiet)


def main(cmd, lstinfo, name, dbpath, min_id, outdir, clust_mode, spe_dir, threads, outfile=None,
         verbose=0, quiet=False):
    """
    Main method, doing all steps:

    - concatenate all protein files
    - create database as ffindex
    - cluster all proteins
    - convert to pangenome file
    - creating summary and matrix of pangenome

    Parameters
    ----------
    lstinfo : str
        file with name of genomes to consider for pan in the first column, without extension.
        Other columns are ignored. The first column header must be 'gembase_name'
    name : str
        name given to the dataset. For example, ESCO44 for 44 *Escherichia coli* genomes.
    dbpath : str
        path to the folder containing all protein files (files called as the name of genome
        given in lstinfo + ".prt"
    min_id : float
        Minimum percentage of identity between 2 proteins to put them in the same family
    outdir : str
        path to folder which will contain pangenome results and tmp files
    clust_mode : [0, 1, 2]
        0 for 'set cover', 1 for 'single-linkage', 2 for 'CD-Hit'
    spe_dir : str or None
        path to the folder where concatenated bank of proteins must be saved.
        None to use the same folder as protein files
    threads : int
        Max number of threads to use
    outfile : str or None
        Name of the pangenome. None to use the default name
    verbose : int
            verbosity:
        - defaut 0 : stdout contains INFO, stderr contains ERROR.
        - 1: stdout contains INFO, stderr contains WARNING and ERROR
        - 2: stdout contains (DEBUG), DETAIL and INFO, stderr contains WARNING and ERROR
        - >=15: Add DEBUG in stdout
    quiet : bool
        True if nothing must be sent to stdout/stderr, False otherwise
    """
    # import needed packages
    import logging
    from PanACoTA import utils
    from PanACoTA.pangenome_module import protein_seq_functions as protf
    from PanACoTA.pangenome_module import mmseqs_functions as mmf
    from PanACoTA.pangenome_module import post_treatment as pt
    from PanACoTA import __version__ as version

    # test if mmseqs is installed and in the path
    if not utils.check_installed("mmseqs"):  # pragma: no cover
        print("mmseqs is not installed. 'PanACoTA pangenome' cannot run.")
        sys.exit(1)

    os.makedirs(outdir, exist_ok=True)
    # level is the minimum level that will be considered.
    # for verbose = 0 or 1, ignore details and debug, start from info
    if verbose <= 1:
        level = logging.INFO
    # for verbose = 2, ignore only debug
    if verbose >= 2 and verbose < 15:
        level = utils.detail_lvl() # int corresponding to detail level
    # for verbose >= 15, write everything
    if verbose >= 15:
        level = logging.DEBUG
    # name logfile, add timestamp if already existing
    logfile_base = os.path.join(outdir, "PanACoTA-pangenome_" + name)
    utils.init_logger(logfile_base, level, '', verbose=verbose, quiet=quiet, log_details=True)
    logger = logging.getLogger("pangenome")
    logger.info(f'PanACoTA version {version}')
    logger.info("Command used\n \t > " + cmd)

    # Build bank with all proteins to include in the pangenome
    logger.info("build prt bank")
    prt_path = protf.build_prt_bank(lstinfo, dbpath, name, spe_dir, quiet)
    # Do pangenome
    families, panfile = mmf.run_all_pangenome(min_id, clust_mode, outdir,
                                              prt_path, threads, outfile, quiet)
    # Create matrix pan_quali, pan_quanti and summary file
    pt.post_treat(families, panfile)
    logger.info("DONE")
    return panfile


def build_parser(parser):
    """
    Method to create a parser for command-line options

    Parameters
    ----------
    parser : argparse.ArgumentParser
        parser to configure in order to extract command-line arguments
    """
    import argparse
    from PanACoTA import utils_argparse

    # Create command-line parser for all options and arguments to give
    required = parser.add_argument_group('Required arguments')
    required.add_argument("-l", dest="lstinfo_file", required=True,
                          help=("File containing the list of all genomes to include in "
                                "the pan-genome, 1 genome per line: it can be the "
                                "LSTINFO-<list_file>.lst file of 'PanACoTA annotate' module."
                                "Here, only the first column (genome name without extension) "
                                "will be used. All proteins of all these genomes will be "
                                "concatenated in a file called <dataset_name>.All.prt. The "
                                "column header must be 'gembase_name'."))
    required.add_argument("-n", dest="dataset_name", required=True,
                          help=("Name of the dataset which will be clustered (for example, "
                                "SAEN1234 for 1234 Salmonella enterica genomes). This name will "
                                "be used to name the protein databank, a well as the "
                                "pangenome files."))
    required.add_argument("-d", dest="dbpath", required=True,
                          help=("Path to the folder containing all protein files corresponding "
                                "to the genomes of the dataset (output directory 'Proteins' "
                                "of 'PanACoTA annotate' module)."))
    required.add_argument("-o", dest="outdir", required=True,
                          help=("Output directory, where all results must be saved "
                                "(including tmp folder)"))

    optional = parser.add_argument_group('Optional arguments')
    optional.add_argument("-i", dest="min_id", type=utils_argparse.perc_id, default=0.8,
                          help=("Minimum sequence identity to be considered in the same "
                                "cluster (float between 0 and 1). Default is 0.8."))
    optional.add_argument("-f", dest="outfile",
                          help=("Use this option if you want to give the name of the pangenome "
                                "output file (without path). Otherwise, by default, it is called "
                                "PanGenome-mmseq_<given_dataset_name>.All.prt_<"
                                "information_on_parameters>.lst"))
    optional.add_argument("-c", dest="clust_mode", type=int, choices=[0, 1, 2], default=1,
                          help=("Choose the clustering mode: 0 for 'set cover', 1 for "
                                "'single-linkage', 2 for 'CD-Hit'. Default "
                                "is 'single-linkage' (1)"))
    optional.add_argument("-s", dest="spedir",
                          help=("use this option if you want to save the concatenated protein "
                                "databank in another directory than the one containing all "
                                "individual protein files ('Proteins' folder)."))
    optional.add_argument("--threads", dest="threads", default=1, type=utils_argparse.thread_num,
                          help=("add this option if you want to parallelize on several threads. "
                                "Indicate on how many threads you want to parallelize. "
                                "By default, it uses 1 thread. Put 0 if you want to use "
                                "all threads of your computer."))

    helper = parser.add_argument_group('Others')
    helper.add_argument("-v", "--verbose", dest="verbose", action="count", default=0,
                        help="Increase verbosity in stdout/stderr.")
    helper.add_argument("-q", "--quiet", dest="quiet", action="store_true", default=False,
                        help=("Do not display anything to stdout/stderr. log files will "
                              "still be created."))
    helper.add_argument("-h", "--help", dest="help", action="help",
                        help="show this help message and exit")


def parse(parser, argu):
    """
    Parse arguments given to parser

    Parameters
    ----------
    parser : argparse.ArgumentParser
        Parser to use to parse command-line arguments
    argu : [str]
        command-line given

    Returns
    -------
    argparse.Namespace or None
        The arguments parsed, updated according to some rules. Exit program
        with error message if error occurs with arguments given.
    """
    args = parser.parse_args(argu)
    return args


if __name__ == '__main__':
    import argparse

    my_parser = argparse.ArgumentParser(description="Compute pan-genome", add_help=False)
    build_parser(my_parser)
    OPTIONS = parse(my_parser, sys.argv[1:])
    main_from_parse(OPTIONS)



#!/usr/bin/env python3
# coding: utf-8

"""
Functional tests for the parser of 'annotate' subcommand
"""
import pytest
import argparse
import time
import os
import shutil

from PanACoTA.subcommands import annotate as annot
DBDIR = os.path.join("test", "data", "annotate")
GENEPATH = os.path.join(DBDIR, "generated_by_unit-tests")


@pytest.fixture(autouse=True)
def setup_teardown_module():
    """
    Remove log files at the end of this test module

    Before each test:
    - init logger
    - create directory to put generated files

    After:
    - remove all log files
    - remove directory with generated results
    """
    os.mkdir(GENEPATH)
    print("setup")

    yield
    shutil.rmtree(GENEPATH)
    print("teardown")


def test_parser_noarg(capsys):
    """
    Test that when the script is called without any argument, an error message appears,
    indicating the required arguments.
    """
    parser = argparse.ArgumentParser(description="Annotate all genomes", add_help=False)
    annot.build_parser(parser)
    with pytest.raises(SystemExit):
        annot.parse(parser, "".split())
    _, err = capsys.readouterr()
    assert "[-d DB_PATH] -r RES_PATH [-l LIST_FILE] [-n NAME] [-Q]" in err
    assert "[--info FROM_INFO] [--prodigal] [--small] [--l90 L90]" in err
    assert "[--nbcont NBCONT] [--cutn CUTN] [--date DATE] [--tmp TMPDIR]" in err
    assert "[--annot_dir ANNOTDIR] [-F] [--threads THREADS] [-v]" in err
    assert "[-q] [-h]" in err
    assert "the following arguments are required: -r" in err


def test_parser_noname(capsys):
    """
    Test that when the script is called without any name for the genomes not -Q option,
    it returns an error message
    """
    parser = argparse.ArgumentParser(description="Annotate all genomes", add_help=False)
    annot.build_parser(parser)
    with pytest.raises(SystemExit):
        annot.parse(parser, "-r respath".split())
    _, err = capsys.readouterr()
    assert ("You must specify your genomes dataset name in 4 characters with "
            "'-n name' option (type -h for more information). Or, if you do not want "
            "to annotate and format your genomes but just to run quality control, use "
            "option '-Q") in err


# def test_parser_wrongname(capsys):
#     """
#     Test that when the script is called with a genome name with more than 4 characters,
#     it returns an error message
#     """
#     parser = argparse.ArgumentParser(description="Annotate all genomes", add_help=False)
#     annot.build_parser(parser)
#     with pytest.raises(SystemExit):
#         annot.parse(parser, "list_file -d dbpath -r respath -n genome".split())
#     _, err = capsys.readouterr()
#     assert ("The genome name must contain 4 characters. For example, this name can "
#             "correspond to the 2 first letters of genus, and 2 first letters of "
#             "species, e.g. ESCO for Escherichia Coli.") in err


def test_parser_negative_cont(capsys):
    """
    Test that when the script is called with a limit of contig number <0,
    it returns an error message
    """
    parser = argparse.ArgumentParser(description="Annotate all genomes", add_help=False)
    annot.build_parser(parser)
    with pytest.raises(SystemExit):
        annot.parse(parser, "-d dbpath -r respath -n g123 --nbcont -5".split())
    _, err = capsys.readouterr()
    assert "The maximum number of contigs allowed must be a positive number." in err


def test_parser_high_cont(capsys):
    """
    Test that when the script is called with a negative limit of contig number,
    it returns an error message
    """
    parser = argparse.ArgumentParser(description="Annotate all genomes", add_help=False)
    annot.build_parser(parser)
    with pytest.raises(SystemExit):
        annot.parse(parser, "-Q -r respath -n g123 --nbcont 10005".split())
    _, err = capsys.readouterr()
    assert "We do not support genomes with more than 9999 contigs." in err


def test_parser_wrong_cont(capsys):
    """
    Test that when the script is called with a non integer limit of contig number,
    it returns an error message
    """
    parser = argparse.ArgumentParser(description="Annotate all genomes", add_help=False)
    annot.build_parser(parser)
    with pytest.raises(SystemExit):
        annot.parse(parser, "-Q -r respath -n g123 --nbcont 10.5".split())
    _, err = capsys.readouterr()
    assert "argument --nbcont: invalid int value: 10.5" in err


def test_parser_wrongl90(capsys):
    """
    Test that when the user does not give an int for the l90 limit, it returns an
    error message.
    """
    parser = argparse.ArgumentParser(description="Annotate all genomes", add_help=False)
    annot.build_parser(parser)
    with pytest.raises(SystemExit):
        annot.parse(parser, "-n TOTO -r respath -n g123 --l90 l90".split())
    _, err = capsys.readouterr()
    assert "argument --l90: invalid int value: 'l90'" in err


def test_parser_wrong_cut(capsys):
    """
    Test that when the user does not give an int for the cutN value, it returns an
    error message.
    """
    parser = argparse.ArgumentParser(description="Annotate all genomes", add_help=False)
    annot.build_parser(parser)
    with pytest.raises(SystemExit):
        annot.parse(parser, "-Q -d dbpath -r respath -n g123 --cutn 10.5".split())
    _, err = capsys.readouterr()
    assert "argument --cutn: invalid int value: '10.5'" in err


def test_parser_wrong_thread(capsys):
    """
    Test that when the user does not give an int for the threads value, it returns an
    error message.
    """
    parser = argparse.ArgumentParser(description="Annotate all genomes", add_help=False)
    annot.build_parser(parser)
    with pytest.raises(SystemExit):
        annot.parse(parser, "-Q -d dbpath -r respath -n g123 --threads 10.5".split())
    _, err = capsys.readouterr()
    assert "argument --threads threads: invalid int value: 10.5" in err


def test_parser_wrong_date(capsys):
    """
    Test that when the user does not give an int for the threads value, it returns an
    error message.
    """
    parser = argparse.ArgumentParser(description="Annotate all genomes", add_help=False)
    annot.build_parser(parser)
    with pytest.raises(SystemExit):
        annot.parse(parser, "-Q -d dbpath -r respath -n g123 --date 417".split())
    _, err = capsys.readouterr()
    assert ("The date must contain 4 characters. Usually, it contains 4 digits, "
            "corresponding to the month (2 digits) and year (2 digits).") in err


def test_parser_q_and_v(capsys):
    """
    Test that when the user wants both quiet and verbose option, it gives an error message
    """
    parser = argparse.ArgumentParser(description="Annotate all genomes", add_help=False)
    annot.build_parser(parser)
    with pytest.raises(SystemExit):
        annot.parse(parser, "-d dbpath -r respath -n g123 -q -v".split())
    _, err = capsys.readouterr()
    assert ("Choose between a verbose output (-v) or a quiet output (-q). "
            "You cannot have both.") in err


def test_parser_default():
    """
    Test that when run with the minimum required arguments, all default values are
    as expected.
    """
    parser = argparse.ArgumentParser(description="Annotate all genomes", add_help=False)
    annot.build_parser(parser)
    options = annot.parse(parser, "-r respath -n g123 -l list_genomes -d dbpath".split())
    assert options.list_file == "list_genomes"
    assert options.db_path == "dbpath"
    assert options.res_path == "respath"
    assert options.name == "g123"
    assert options.l90 == 100
    assert options.nbcont == 999
    assert options.cutn == 5
    assert options.threads == 1
    assert options.date == time.strftime("%m%y")
    assert not options.force
    assert not options.qc_only
    assert not options.from_info
    assert not options.prodigal_only


def test_parser_values():
    """
    Test that values for L90, nbcontig, cutn, threads, date are taken into account
    """
    parser = argparse.ArgumentParser(description="Annotate all genomes", add_help=False)
    annot.build_parser(parser)
    options = annot.parse(parser, ("-l list_file -d dbpath -r respath -n g123 --l90 2 "
                                   "--nbcont 10 --cutn 0 --threads 2 --date toto "
                                   "--prodigal -F").split())
    assert options.list_file == "list_file"
    assert options.db_path == "dbpath"
    assert options.res_path == "respath"
    assert options.name == "g123"
    assert options.l90 == 2
    assert options.nbcont == 10
    assert options.cutn == 0
    assert options.threads == 2
    assert options.date == "toto"
    assert options.force
    assert not options.qc_only
    assert not options.from_info
    assert options.prodigal_only


def test_parser_wrongforce(capsys):
    """
    Test that when run with '-F' option + a value, it returns an error message.
    """
    parser = argparse.ArgumentParser(description="Annotate all genomes", add_help=False)
    annot.build_parser(parser)
    with pytest.raises(SystemExit):
        annot.parse(parser, "-d dbpath -r respath -n g123 -F 10".split())
    _, err = capsys.readouterr()
    assert "unrecognized arguments: 10" in err


def test_parser_qc():
    """
    Test that when run with '-Q' option (for QC only) and no name given for the genome, it
    is set to "NONE"
    """
    parser = argparse.ArgumentParser(description="Annotate all genomes", add_help=False)
    annot.build_parser(parser)
    options = annot.parse(parser, "-l list_file -d dbpath -r respath -Q".split())
    assert options.list_file == "list_file"
    assert options.db_path == "dbpath"
    assert options.res_path == "respath"
    assert options.name == "NONE"
    assert options.l90 == 100
    assert options.nbcont == 999
    assert options.cutn == 5
    assert options.threads == 1
    assert options.date == time.strftime("%m%y")
    assert not options.force
    assert options.qc_only


def test_parser_info_cutn(capsys):
    """
    Test that when run with --info and --cutn x : error message
    If we run from info file, will not touch the sequences.
    """
    parser = argparse.ArgumentParser(description="Annotate all genomes", add_help=False)
    annot.build_parser(parser)
    with pytest.raises(SystemExit):
        annot.parse(parser, "-l list_file -d dbpath -r respath -n name "
                            "--info infofile --cutn 10".split())
    _, err = capsys.readouterr()
    assert ("If you provide a list of genomes with their calculated L90 and number of contigs, "
            "PanACoTA will use the given sequences as is. It will not cut them. So, you cannot "
            "use both --cutn and --info.") in err

def test_info_and_lstfile(capsys):
    """
    Test that there is an error message if user gives both -l infofile and --info LSTINFO
    """
    parser = argparse.ArgumentParser(description="Annotate all genomes", add_help=False)
    annot.build_parser(parser)
    with pytest.raises(SystemExit):
        annot.parse(parser, "-d dbpath -r respath -n name --nbcont 20 -l toto --info info".split())
    _, err = capsys.readouterr()
    assert ("Either you want to annotate raw sequences (name of files in '-l infofile') "
            "which will first go through the QC process, "
            "OR you already did QC on your sequences and just want to annotate them "
            "(information on those sequences in '--info LSTINFO-file'). "
            "Please choose one of these 2 possibilities.") in err


def test_parser_noinfo_nolist(capsys):
    """
    Test that when run without --info nor -l : error message
    Must provide one of them
    """
    parser = argparse.ArgumentParser(description="Annotate all genomes", add_help=False)
    annot.build_parser(parser)
    with pytest.raises(SystemExit):
        annot.parse(parser, "-d dbpath -r respath -n name --nbcont 20".split())
    _, err = capsys.readouterr()
    assert ("You must provide a list of genomes to annotate. Either raw genomes "
            "(see -l option), or genomes with quality information (see --info option).") in err


def test_parser_noinfo_nodbpath(capsys):
    """
    Test that when run without --info nor -d : error message
    Needs a path to genomes to annotate!
    """
    parser = argparse.ArgumentParser(description="Annotate all genomes", add_help=False)
    annot.build_parser(parser)
    with pytest.raises(SystemExit):
        annot.parse(parser, "-r respath -n name --nbcont 20 -l listgenomes".split())
    _, err = capsys.readouterr()
    assert ("You must provide a path to your database genome sequences (-d <db_path>). "
            "If you already have a LSTINFO file, it contains this db_path. Use it "
            "with --info <lstinfo file> option.") in err


def test_parser_info_dbpath(capsys):
    """
    Test that when run with both --info and -d : error message
    Must know which genomes to annotate between the 2
    """
    parser = argparse.ArgumentParser(description="Annotate all genomes", add_help=False)
    annot.build_parser(parser)
    with pytest.raises(SystemExit):
        annot.parse(parser, "-r respath -n name --info file -d dbpath".split())
    _, err = capsys.readouterr()
    assert ("If you run from your LSTINFO file, this one already contains the path "
            "of genomes to annotate. Remove -d <db_path> option.") in err


def test_parser_small_noprodigal(capsys):
    """
    Test that when run with both --small but do not ask to use prodigal, it returns error
    """
    parser = argparse.ArgumentParser(description="Annotate all genomes", add_help=False)
    annot.build_parser(parser)
    with pytest.raises(SystemExit):
        annot.parse(parser, "-r respath -n name --small".split())
    _, err = capsys.readouterr()
    assert("You cannot use --small option with prokka. "
           "Either use prodigal, or remove this option") in err


def test_parser_filter(capsys):
    """
    Test that warnings are written (when will split l90 and/or nbcont)
    """
    parser = argparse.ArgumentParser(description="Annotate all genomes", add_help=False)
    annot.build_parser(parser)
    options = annot.parse(parser, "-l list_file -d dbpath -r respath -Q --l90 10".split())
    assert options.list_file == "list_file"
    assert options.db_path == "dbpath"
    assert options.res_path == "respath"
    assert options.name == "NONE"
    assert options.l90 == 10
    assert options.nbcont == 999
    assert options.cutn == 5
    assert options.threads == 1
    assert options.date == time.strftime("%m%y")
    assert not options.force
    assert options.qc_only
    stdout, _ = capsys.readouterr()
    assert(" !! Your genomes will be filtered, and only the ones with 'L90' <= 10 and 'number of contigs' < 999 "
           "will be kept. If you want to change those thresholds, use '--l90' and '--nbcont' options.") in stdout
    assert("! Your genomes will be split when sequence contains at least 5'N' in a row. "
           "If you want to change this threshold, see --cutn option.") in stdout



def test_parser_nosplit(capsys):
    """
    Test that warnings are written (when will split l90 and/or nbcont)
    """
    parser = argparse.ArgumentParser(description="Annotate all genomes", add_help=False)
    annot.build_parser(parser)
    options = annot.parse(parser, "--info infofile -r respath -Q".split())
    assert not options.list_file
    assert not options.db_path
    assert options.from_info == "infofile"
    assert options.res_path == "respath"
    assert options.name == "NONE"
    assert options.l90 == 100
    assert options.nbcont == 999
    assert options.cutn == 5
    assert options.threads == 1
    assert options.date == time.strftime("%m%y")
    assert not options.force
    assert options.qc_only
    stdout, _ = capsys.readouterr()
    assert (" !! Your sequences will be used as is by PanACoTA. Be sure you already split your sequences at each row of X 'N' if needed.") in stdout
    assert ("PanACoTA will use the values (L90, nbcont) given in your info file. It will ignore the genomes for which those values are incorrect. It will also ignore genomes with more than 999 contigs.")



#!/usr/bin/env python3
# coding: utf-8

"""
Unit tests for the functions of utils-argparse.py dealing with checking arguments given to argparse
"""

import logging
import os
import pytest
import shutil
import argparse
import multiprocessing

import PanACoTA.utils_argparse as autils

CONFFILE = os.path.join("test", "data", "utils", "configfile.ini")

def test_gen_name():
    """
    Test that, when giving a gene name, if it does not have 4 characters (letters or num),
    it returns an error message
    """
    assert autils.gen_name("TOTO") == "TOTO"
    assert autils.gen_name("1234") == "1234"
    assert autils.gen_name("T1O2") == "T1O2"
    with pytest.raises(argparse.ArgumentTypeError) as err:
        autils.gen_name("a long name")
    assert ("The genome name must contain 4 characters. For example, this name "
            "can correspond to the 2 first letters of genus, and 2 first letters of "
            "species, e.g. ESCO for Escherichia Coli") in str(err.value)
    with pytest.raises(argparse.ArgumentTypeError) as err:
        autils.gen_name("-gdd")
    assert ("The genome name must contain 4 characters. For example, this name can correspond "
            "to the 2 first letters of genus, and 2 first letters of species, e.g. "
            "ESCO for Escherichia Coli") in str(err.value)


def test_date_name(capsys):
    """
    Test that when a date is given, it returns expected message
    """
    assert autils.date_name("0920") == "0920"
    assert autils.date_name("se20") == "se20"
    with pytest.raises(argparse.ArgumentTypeError) as err:
        autils.date_name("september 2020")
    assert ("The date must contain 4 characters. Usually, it contains 4 digits, "
            "corresponding to the month (2 digits) and year (2 digits)") in str(err.value)


def test_get_date():
    """
    test that it returns current date
    """
    import time
    t = time.strftime("%m%y")
    assert autils.get_date() == t


def test_cont_num():
    """
    Test that given value of contig number is valid
    """
    assert autils.cont_num(10) == 10
    with pytest.raises(argparse.ArgumentTypeError) as raised_err:
        autils.cont_num("-2")
    assert ("The maximum number of contigs allowed must be a "
            "positive number") in str(raised_err.value)
    with pytest.raises(argparse.ArgumentTypeError) as err:
        autils.cont_num("10000")
    assert ("We do not support genomes with more than 9999 contigs") in str(err.value)
    with pytest.raises(argparse.ArgumentTypeError) as err:
        autils.cont_num("a")
    assert ("argument --nbcont: invalid int value: a") in str(err.value)
    with pytest.raises(argparse.ArgumentTypeError) as err:
        autils.cont_num("1.1")
    assert ("argument --nbcont: invalid int value: 1.") in str(err.value)


def test_thread_num():
    """
    Test that given number of threads is as expected
    """
    assert autils.thread_num("1") == 1
    with pytest.raises(argparse.ArgumentTypeError) as err:
        a = autils.thread_num("1.1")
    assert ("argument --threads threads: invalid int value: 1.1") in str(err.value)
    with pytest.raises(argparse.ArgumentTypeError) as err:
        a = autils.thread_num("a")
    assert ("argument --threads threads: invalid int value: a") in str(err.value)
    nb_cpu = multiprocessing.cpu_count()
    with pytest.raises(argparse.ArgumentTypeError) as err:
        a = autils.thread_num(str(nb_cpu*2))
    assert (f"You have {nb_cpu} threads on your computer, you cannot ask for more: invalid value: "
            f"{nb_cpu*2}") in str(err.value)
    with pytest.raises(argparse.ArgumentTypeError) as err:
        a = autils.thread_num("-1")
    assert ("Please provide a positive number of threads (or 0 for all threads): "
            "Invalid value: -1") in str(err.value)
    assert autils.thread_num(0) == nb_cpu


def test_positive_int():
    """
    Test checking that given argument is a positive integer
    """
    assert autils.positive_int("1") == 1
    with pytest.raises(argparse.ArgumentTypeError) as err:
        a = autils.positive_int("1.1")
    assert ("argument --cutn: invalid int value: '1.1'") in str(err.value)
    with pytest.raises(argparse.ArgumentTypeError) as err:
        a = autils.positive_int("-1")
    assert ("error: argument --cutn must be a positive integer: "
            "invalid int value: '-1'") in str(err.value)


def test_mash_dist():
    """
    Test checking that given value is ok for a mash distance
    """
    assert autils.mash_dist("0.05") == 0.05
    assert autils.mash_dist("1e-4") == 0.0001
    with pytest.raises(argparse.ArgumentTypeError) as err:
        a = autils.mash_dist("1.1.1")
    assert ("error: mash distance: invalid float value: '1.1.1'") in str(err.value)
    with pytest.raises(argparse.ArgumentTypeError) as err:
        a = autils.mash_dist("one")
    assert ("error: mash distance: invalid float value: 'one'") in str(err.value)
    with pytest.raises(argparse.ArgumentTypeError) as err:
        a = autils.mash_dist("1.000001")
    assert ("error: mash distance must be between 0 and 1: "
            "invalid value: '1.000001'") in str(err.value)
    with pytest.raises(argparse.ArgumentTypeError) as err:
        a = autils.mash_dist("-1e-4")
    assert ("error: mash distance must be between 0 and 1: "
            "invalid value: '-0.0001'") in str(err.value)


def test_percentage():
    """
    Test value given for parameter -t tol:
    - string corresponding to a float or an int -> returns float value
    - a non digital param for a percentage raises appropriate error
    - a negative number -> appropriate error
    - number>1 -> appropriate error
    """
    assert autils.percentage("0.5") == 0.5
    assert type(autils.percentage("0.5")) == float
    assert autils.percentage("1e-1") == 0.1
    assert autils.percentage("1") == 1.0
    with pytest.raises(argparse.ArgumentTypeError) as err:
        a = autils.percentage("one")
    assert ("argument -t tol: invalid float value: one") in str(err.value)
    with pytest.raises(argparse.ArgumentTypeError) as err:
        a = autils.percentage(-0.5)
    assert ("The minimum %% of genomes required in a family to be persistent must "
            "be in [0, 1]. Invalid value: -0.5") in str(err.value)
    with pytest.raises(argparse.ArgumentTypeError) as err:
        a = autils.percentage("1.1")
    assert ("The minimum %% of genomes required in a family to be persistent must "
            "be in [0, 1]. Invalid value: 1.1") in str(err.value)


def test_perc_id():
    """
    Same as test_percentage, but for parameter -i percentage_id
    """
    assert autils.perc_id("0.5") == 0.5
    assert type(autils.perc_id("0.5")) == float
    assert autils.perc_id("1e-1") == 0.1
    assert autils.perc_id("1") == 1.0
    with pytest.raises(argparse.ArgumentTypeError) as err:
        a = autils.perc_id("one")
    assert ("argument -i percentage_id: invalid float value: one") in str(err.value)
    with pytest.raises(argparse.ArgumentTypeError) as err:
        a = autils.perc_id(-0.5)
    assert ("The minimum %% of identity must be in [0, 1]. Invalid value: -0.5") in str(err.value)
    with pytest.raises(argparse.ArgumentTypeError) as err:
        a = autils.perc_id("1.1")
    assert ("The minimum %% of identity must be in [0, 1]. Invalid value: 1.1") in str(err.value)


def test_conf_parser_init_empty(capsys):
    """
    test class Conf_all_parser init when no config file or empty config file
    """
    # No conf file, no section
    c = autils.Conf_all_parser("")
    assert c.sec_dicts == {}
    # No conf file, empty section list
    c = autils.Conf_all_parser("", [])
    assert c.sec_dicts == {}
    # No conf file, sections
    c = autils.Conf_all_parser("", ["sec1", "sec2"])
    assert c.sec_dicts == {"sec1": {}, "sec2": {}}

    confdir = os.path.join("test", "data", "generated_by_utils")
    os.makedirs(confdir)
    conffile = os.path.join(confdir, "conf.ini")
    open(conffile, "w").close()
    # Empty conffile, no section
    c = autils.Conf_all_parser(conffile)
    assert c.sec_dicts == {}
    # Empty conffile, sections
    c = autils.Conf_all_parser(conffile, ["sec1", "sec2"])
    assert c.sec_dicts == {"sec1": {}, "sec2": {}}
    shutil.rmtree(confdir)


def test_conf_parser_init():
    """
    Test config parser with a given config file. Check value of defaults etc.
    """
    # configfile but no section
    c = autils.Conf_all_parser(CONFFILE)
    assert c["sec1"]["toto"] == "parameter"
    assert c["sec1"]["param1"] == "10"
    assert c["sec2"]["param1"] == "3"
    assert c["sec3"]["param1"] == "3"
    assert c["sec1"]["param2"] == "10"
    assert c["sec2"]["param2"] == ''
    assert c["sec3"]["param2"] == "10"
    assert c.sec_dicts == {}

    # configfile and 2 sections among 3 existing in configfile
    c = autils.Conf_all_parser(CONFFILE, ["sec1", "sec3"])
    assert c["sec1"]["toto"] == "parameter"
    assert c["sec1"]["param1"] == "10"
    assert c["sec2"]["param1"] == "3"
    assert c["sec2"]["myval"] == "parameter"
    assert c["sec3"]["param1"] == "3"
    assert c["sec1"]["param2"] == "10"
    assert c["sec2"]["param2"] == ''
    assert c["sec3"]["param2"] == "10"
    assert c.sec_dicts == {"sec1": {"param1": "10", "param2": "10", 'sec1p': "",
                                    "toto": "parameter", "titi": "my value"},
                           "sec3": {"param1": "3", "param2": "10"}}

    # configfile 2 sections, 1 not in configfile
    c = autils.Conf_all_parser(CONFFILE, ["sec1", "sec4"])
    assert c["sec1"]["toto"] == "parameter"
    assert c["sec1"]["param1"] == "10"
    assert c["sec2"]["param1"] == "3"
    assert c["sec4"]["param1"] == "3"  # created sec4 with default parameters
    assert c["sec4"]["param2"] == "10"  # created sec4 with default parameters
    assert not "toto" in c["sec4"]  # created sec4 with default parameters
    assert c.sec_dicts == {"sec1": {"param1": "10", "param2": "10", 'sec1p': "",
                                    "toto": "parameter", "titi": "my value"},
                           "sec4": {}}  # But sec4 dict is empty (no param given in configfile)


def test_conf_parser_get_section(capsys):
    """
    Test get dict of values for a given section
    """
    c = autils.Conf_all_parser(CONFFILE, ["sec1", "sec4"])
    # Get sec1 dict
    assert c.get_section_dict("sec1") == {"param1": "10", "param2": "10", 'sec1p': "",
                                          "toto": "parameter", "titi": "my value"}
    assert c.get_section_dict("sec4") == {}
    # Try to get sec2 dict, but does not exist
    with pytest.raises(SystemExit):
        c.get_section_dict("sec2")
    out, err = capsys.readouterr()
    assert ('No section sec2 in test/data/utils/configfile.ini') in out


def test_conf_parser_add_default(capsys):
    """
    Test add default parameters to config parser.
    If parameter given already exists, do nothing. If does not exist, add it with given value
    """
    c = autils.Conf_all_parser(CONFFILE, ["sec1", "sec4"])
    defaults = {"param1": "55", "defparam": "default value"}
    # Add default parameters to sec1
    c.add_default(defaults, "sec1")
    assert c.get_section_dict("sec1") == {"param1": "10", "param2": "10", 'sec1p': "",
                                          "toto": "parameter", "titi": "my value",
                                          "defparam": "default value"}
    assert c.get_section_dict("sec4") == {}
    c["sec1"]["param1"] == "10"
    c["sec1"]["defparam"] == "default value"
    c["sec4"]["param1"] == "3"   # sec4 has default parameter found in configfile

    # Add default parameters to sec4 (sec1 already have them)
    c.add_default(defaults, "sec4")
    assert c.get_section_dict("sec1") == {"param1": "10", "param2": "10", 'sec1p': "",
                                          "toto": "parameter", "titi": "my value",
                                          "defparam": "default value"}
    assert c.get_section_dict("sec4") == {"param1": "55", "defparam": "default value"}
    c["sec1"]["param1"] == "10"
    c["sec1"]["defparam"] == "default value"
    c["sec4"]["param1"] == "55"


def test_conf_parser_update(capsys):
    """
    Test update section
    Update current parameters with values given
    """
    c1 = autils.Conf_all_parser(CONFFILE, ["sec1", "sec4"])
    update = {"param1": "55", "defparam": "default value"}
    # Add default parameters to sec1
    c1.update(update, "sec1")
    assert c1.get_section_dict("sec1") == {"param1": "55", "param2": "10", 'sec1p': "",
                                          "toto": "parameter", "titi": "my value",
                                          "defparam": "default value"}
    assert c1.get_section_dict("sec4") == {}
    c1["sec1"]["param1"] == "10"
    c1["sec1"]["defparam"] == "default value"
    c1["sec4"]["param1"] == "3"   # sec4 has default parameter found in configfile

    # Add default parameters to sec1
    c2 = autils.Conf_all_parser(CONFFILE, ["sec1", "sec4"])
    c2.update(update, "sec4")
    assert c2.get_section_dict("sec1") == {"param1": "10", "param2": "10", 'sec1p': "",
                                          "toto": "parameter", "titi": "my value"}
    assert c2.get_section_dict("sec4") == {"param1": "55", "defparam": "default value"}
    c2["sec1"]["param1"] == "10"
    c2["sec4"]["param1"] == "55"   # sec4 has default parameter found in configfile


def test_conf_parser_setbool(capsys):
    """
    try to convert a given parameter to a boolean
    """
    c1 = autils.Conf_all_parser(CONFFILE, ["sec_bool"])
    # 0/1 to False/True
    c1.set_boolean("sec_bool", "bool num_false")
    assert c1["sec_bool"]["bool num_false"] == "0"
    assert c1.sec_dicts["sec_bool"]["bool num_false"] == False
    c1.set_boolean("sec_bool", "bool_num_true")
    assert c1["sec_bool"]["bool_num_true"] == "1"
    assert c1.sec_dicts["sec_bool"]["bool_num_true"] == True

    # off/on to False/True
    c1.set_boolean("sec_bool", "bool_f")
    assert c1["sec_bool"]["bool_f"] == "off"
    assert c1.sec_dicts["sec_bool"]["bool_f"] == False
    c1.set_boolean("sec_bool", "bool_t")
    assert c1["sec_bool"]["bool_t"] == "ON"
    assert c1.sec_dicts["sec_bool"]["bool_t"] == True

    # no/yes to False/True
    c1.set_boolean("sec_bool", "bool_n")
    assert c1["sec_bool"]["bool_n"] == "no"
    assert c1.sec_dicts["sec_bool"]["bool_n"] == False
    c1.set_boolean("sec_bool", "bool_y")
    assert c1["sec_bool"]["bool_y"] == "YES"
    assert c1.sec_dicts["sec_bool"]["bool_y"] == True

    # false/true to False/True
    c1.set_boolean("sec_bool", "bool_false")
    assert c1["sec_bool"]["bool_false"] == "FalSe"
    assert c1.sec_dicts["sec_bool"]["bool_false"] == False
    c1.set_boolean("sec_bool", "bool_true")
    assert c1["sec_bool"]["bool_true"] == "tRUE"
    assert c1.sec_dicts["sec_bool"]["bool_true"] == True

    # error
    with pytest.raises(SystemExit):
        c1.set_boolean('sec1', "param1")
    out, err = capsys.readouterr()
    assert ('ERROR: param1 must be a boolean. Wrong value: 10.') in out


def test_conf_parser_setint(capsys):
    """
    try to convert a given parameter to an int
    """
    c1 = autils.Conf_all_parser(CONFFILE, ["sec1"])
    c1.set_int("sec1", "param1")
    assert c1["sec1"]["param1"] == "10"
    assert c1.sec_dicts["sec1"]["param1"] == 10

    with pytest.raises(SystemExit):
        c1.set_int("sec1", "toto")
    out, err = capsys.readouterr()
    assert ('ERROR: toto must be an int. Wrong value: parameter.') in out


def test_conf_parser_setfloat(capsys):
    """
    try to convert a given parameter to a float
    """
    c1 = autils.Conf_all_parser(CONFFILE, ["sec_float"])
    # float with decimal
    c1.set_float("sec_float", "float_param")
    assert c1["sec_float"]["float_param"] == "0.12"
    assert c1.sec_dicts["sec_float"]["float_param"] == 0.12

    # exp format
    c1.set_float("sec_float", "float_param2")
    assert c1["sec_float"]["float_param2"] == "1e5"
    assert c1.sec_dicts["sec_float"]["float_param2"] == 100000

    with pytest.raises(SystemExit):
        c1.set_float("sec1", "toto")
    out, err = capsys.readouterr()
    assert ('ERROR: toto must be a float. Wrong value: parameter.') in out



#!/usr/bin/env python3
# coding: utf-8

"""
Unit tests for the get_seqs submodule in align module
"""
import os
import shutil
import logging

import pytest

import PanACoTA.align_module.get_seqs as gseq
from PanACoTA import utils
import test.test_unit.utilities_for_tests as tutil

# Define common variables
ALDIR = os.path.join("test", "data", "align")
EXPPATH = os.path.join(ALDIR, "exp_files")
TESTPATH = os.path.join(ALDIR, "test_files")
GENEPATH = os.path.join(ALDIR, "generated_by_unit-tests")
DBPATH = os.path.join("test", "data", "pangenome", "test_files", "example_db")
FASTA = os.path.join(DBPATH, "Proteins", "GEN2.1017.00001.prt")
LOGFILE_BASE = "test_getseqs-logs"
LOG_LEVEL = logging.DEBUG
LOGFILES = [LOGFILE_BASE + ext for ext in [".log", ".log.debug", ".log.details", ".log.err"]]


@pytest.fixture(autouse=True)
def setup_teardown_module():
    """
    Remove log files at the end of this test module

    Before each test:
    - init logger
    - create directory to put generated files

    After:
    - remove all log files
    - remove directory with generated results
    """
    utils.init_logger(LOGFILE_BASE, LOG_LEVEL, 'test_getseq', verbose=1)
    os.mkdir(GENEPATH)
    print("setup")

    yield
    for f in LOGFILES:
        if os.path.exists(f):
            os.remove(f)
    shutil.rmtree(GENEPATH)
    print("teardown")


# Start tests
def test_extract_noseq_out_given():
    """
    Test that when giving an open fasta file, an empty list of sequences to extract, and
    an open file to which extracted sequences must be written, the output file is empty.
    """
    to_extract = {}
    outfile = os.path.join(GENEPATH, "test_noextract_out-given.prt")
    with open(FASTA, "r") as fasf, open(outfile, "w") as outf:
        gseq.extract_sequences(to_extract, fasf, outf=outf)
    with open(outfile, "r") as outf:
        assert outf.readlines() == []


def test_extract_seq_out_given():
    """
    Test that when giving an open fasta file, a list of 3 sequences to extract, and
    a open file to which extracted sequences must be written, it writes the
    expected sequences to output file.
    """
    to_extract = ["GEN2.1017.00001.b0001_00001", "GEN2.1017.00001.i0003_00008",
                  "GEN2.1017.00001.b0004_00013"]
    outfile = os.path.join(GENEPATH, "test_extract_out-given.prt")
    with open(FASTA, "r") as fasf, open(outfile, "w") as outf:
        gseq.extract_sequences(to_extract, fasf, outf=outf)
    exp_extracted = os.path.join(EXPPATH, "exp_extracted.prt")
    assert tutil.compare_file_content(outfile, exp_extracted)


def test_extract_seq_out_given_wrongname():
    """
    Test that when giving an open fasta file, a list of 4 sequences to extract, and
    a open file to which extracted sequences must be written, it writes the
    expected sequences to output file. If a sequence does not exist, it is just not written.
    """
    to_extract = {"GEN2.1017.00001.b0001_00001": "", "GEN2.1017.00001.i0003_00008": "",
                  "GEN2.1017.00001.b0004_00013": "", "toto": ""}
    outfile = os.path.join(GENEPATH, "test_extract_out-given.prt")
    with open(FASTA, "r") as fasf, open(outfile, "w") as outf:
        gseq.extract_sequences(to_extract, fasf, outf=outf)
    exp_extracted = os.path.join(EXPPATH, "exp_extracted.prt")
    assert tutil.compare_file_content(outfile, exp_extracted)


def test_extract_seq_out_allsame():
    """
    Test that when giving an open fasta file, a list of 3 sequences to extract with a
    corresponding output file (same for all 3 proteins), it writes the
    expected sequences to output file.
    """
    out = os.path.join(GENEPATH, "test_extract1.prt")
    to_extract = {"GEN2.1017.00001.b0001_00001": out,
                  "GEN2.1017.00001.i0003_00008": out,
                  "GEN2.1017.00001.b0004_00013": out}
    with open(FASTA, "r") as fasf:
        gseq.extract_sequences(to_extract, fasf, files_todo=[out])
    exp_extracted = os.path.join(EXPPATH, "exp_extracted.prt")
    assert tutil.compare_file_content(out, exp_extracted)


def test_extract_seq_out_different():
    """
    Test that when giving an open fasta file, a list of 3 sequences to extract with a
    corresponding output file for each, it writes the expected sequences to the expected output
    file.
    """
    out1 = os.path.join(GENEPATH, "test_extract1.prt")
    out2 = os.path.join(GENEPATH, "test_extract2.prt")
    to_extract = {"GEN2.1017.00001.b0001_00001": out1,
                  "GEN2.1017.00001.i0003_00008": out2,
                  "GEN2.1017.00001.b0004_00013": out1}
    with open(FASTA, "r") as fasf:
        gseq.extract_sequences(to_extract, fasf, files_todo=[out1, out2])
    exp_extracted1 = os.path.join(EXPPATH, "exp_extracted1.prt")
    exp_extracted2 = os.path.join(EXPPATH, "exp_extracted2.prt")
    assert tutil.compare_file_content(out1, exp_extracted1)
    assert tutil.compare_file_content(out2, exp_extracted2)


def test_extract_seq_out_different_notasked():
    """
    Test that when giving an open fasta file, a list of 3 sequences to extract with a
    corresponding output file for each, it writes the
    expected sequences to expected output file, only if this file is contained in 'files_todo'.
    If not, file is not created, and protein not extracted.
    """
    out1 = os.path.join(GENEPATH, "test_extract1.prt")
    out2 = os.path.join(GENEPATH, "test_extract2.prt")
    to_extract = {"GEN2.1017.00001.b0001_00001": out1,
                  "GEN2.1017.00001.i0003_00008": out2,
                  "GEN2.1017.00001.b0004_00013": out1}
    with open(FASTA, "r") as fasf:
        gseq.extract_sequences(to_extract, fasf, files_todo=[out1])
    exp_extracted1 = os.path.join(EXPPATH, "exp_extracted1.prt")
    assert tutil.compare_file_content(out1, exp_extracted1)
    assert not os.path.isfile(out2)


def test_get_names_files():
    """
    Test that given an open tab file (containing 2 columns: name of sequence to extract,
    file where it must be extracted), it returns the expected dict of sequences to extract
    """
    tabfile = os.path.join(EXPPATH, "exp_getentry-gen-ESCO2_write.txt")
    with open(tabfile, "r") as tabf:
        toext = gseq.get_names_to_extract(tabf, outfile=None)
    exp_toext = {'ESCO2_00001': "Aldir/TEST6-current.1.gen",
                 'ESCO2_22': 'Aldir/TEST6-current.2.gen',
                 'ESCO2_456': 'Aldir/TEST6-current.4.gen',
                 'ESCO2_46': 'Aldir/TEST6-current.3.gen'}
    assert toext == exp_toext


def test_get_names_wrongformat(caplog):
    """
    Test that given an open tab file, containing only the sequence name (no file associated) fir
    1 sequence, it exits with the expected error message
    """
    tabfile = os.path.join(TESTPATH, "wrong_getentry.prt")
    with open(tabfile, "r") as tabf:
        with pytest.raises(SystemExit):
            gseq.get_names_to_extract(tabf, outfile=None)
    assert ("Your file test/data/align/test_files/wrong_getentry.prt does not contain an output "
            "filename for ESCO2_22. Please give an output filename for each sequence to "
            "extract, or give a general output filename where all sequences will be "
            "extracted.") in caplog.text


def test_get_names_out_and2columns():
    """
    Test that given an open tab file (containing 2 columns: name of sequence to extract,
    file where it must be extracted), and an output file name, it returns the expected dict of
    sequences to extract (all to extract to the same given output file, the 2nd column of tab
    file is ignored).
    """
    tabfile = os.path.join(EXPPATH, "exp_getentry-gen-ESCO2_write.txt")
    outfile = os.path.join(GENEPATH, "test_getnames_2columns")
    with open(tabfile, "r") as tabf:
        toext = gseq.get_names_to_extract(tabf, outfile=outfile)
    exp_toext = {'ESCO2_00001': outfile,
                 'ESCO2_22': outfile,
                 'ESCO2_456': outfile,
                 'ESCO2_46': outfile}
    assert toext == exp_toext


def test_get_names_out_and1column():
    """
    Test that given an open tab file (containing only sequence names), and an output file name,
    it returns the expected dict of sequences to extract (all to extract to the same given
    output file).
    """
    tabfile = os.path.join(TESTPATH, "getentry_1column.txt")
    outfile = os.path.join(GENEPATH, "test_getnames_1column")
    with open(tabfile, "r") as tabf:
        toext = gseq.get_names_to_extract(tabf, outfile=outfile)
    exp_toext = {'ESCO2_00001': outfile,
                 'ESCO2_22': outfile,
                 'ESCO2_456': outfile,
                 'ESCO2_46': outfile}
    assert toext == exp_toext


def test_get_genome_all_seqs():
    """
    Test that given a fasta file, and a tab file containing all sequences to extract, with the
    files to which it must be extracted, it extracts everything in the right file.
    """
    tabfile = os.path.join(TESTPATH, "getentry_all_2columns.txt")
    todo = ["file1.txt", "file2.txt"]
    todo = [os.path.join(GENEPATH, f) for f in todo]
    gseq.get_genome_seqs(FASTA, tabfile, todo)
    for i in range(1, 3):
        outfile = os.path.join(GENEPATH, f"file{i}.txt")
        exp_file = os.path.join(EXPPATH, f"exp_extracted{i}.prt")
        assert os.path.isfile(outfile)
        assert tutil.compare_file_content(outfile, exp_file)


def test_get_genome_seqs_outgiven_2cols():
    """
    Test that given a fasta file, and a tab file containing all sequences to extract, with the
    files to which it must be extracted, and an output file, it extracts all sequences to the
    same output file, ignoring the ones given in tab file
    """
    tabfile = os.path.join(TESTPATH, "getentry_all_2columns.txt")
    outfile = os.path.join(GENEPATH, "fileout.txt")
    todo = []
    gseq.get_genome_seqs(FASTA, tabfile, todo, outfile)
    assert os.path.isfile(outfile)
    exp_file = os.path.join(EXPPATH, "exp_extracted.prt")
    assert tutil.compare_file_content(outfile, exp_file)
    os.remove(outfile)


def test_get_genome_seqs_outgiven_1col():
    """
    Test that given a fasta file, and a tab file containing only all sequences to extract,
    (no filename), and an output file, it extracts all sequences to the same output file.
    """
    tabfile = os.path.join(TESTPATH, "getentry_all_1column.txt")
    outfile = os.path.join(GENEPATH, "fileout.txt")
    todo = []
    gseq.get_genome_seqs(FASTA, tabfile, todo, outfile)
    assert os.path.isfile(outfile)
    exp_file = os.path.join(EXPPATH, "exp_extracted.prt")
    assert tutil.compare_file_content(outfile, exp_file)


def test_get_genome_seqs_1notasked():
    """
    Test that given a fasta file, and a tab file containing all sequences to extract, with the
    files to which it must be extracted, and only 1 of them in 'files_todo', it extracts only
    the proteins going to this file.
    """
    tabfile = os.path.join(TESTPATH, "getentry_all_2columns.txt")
    outfile1 = os.path.join(GENEPATH, "file1.txt")
    outfile2 = os.path.join(GENEPATH, "file2.txt")
    todo = [outfile1]
    gseq.get_genome_seqs(FASTA, tabfile, todo)
    assert os.path.isfile(outfile1)
    exp_file = os.path.join(EXPPATH, "exp_extracted1.prt")
    assert tutil.compare_file_content(outfile1, exp_file)
    assert not os.path.isfile(outfile2)


def test_get_genome_seqs_exists(caplog):
    """
    Test that when the output file given aleady exists, it does not overwrite it, but returns a
    warning message saying that it will be used for next step.
    """
    tabfile = os.path.join(TESTPATH, "getentry_all_2columns.txt")
    outfile = os.path.join(GENEPATH, "fileout.txt")
    open(outfile, "w").close()
    todo = []
    gseq.get_genome_seqs(FASTA, tabfile, todo, outfile)
    assert ("Sequences are already extracted in test/data/align/generated_by_unit-tests/fileout.txt. "
            "This will be used for next step. If "
            "you want to re-extract all sequences, use option -F (or --force)") in caplog.text
    with open(outfile, "r") as outf:
        assert outf.readlines() == []


def test_check_extract_empty():
    """
    Test that when no file is present, it returns all gen and prt files, for all families
    """
    all_fams = [1, 6, 10, 11, 13, 14]
    aldir = os.path.join(GENEPATH, "Aldir")
    dname = "TESTcheckExtract"
    files_todo = gseq.check_existing_extract(all_fams, aldir, dname)
    assert len(files_todo) == 12
    gens = [os.path.join(aldir, f"{dname}-current.{num}.gen") for num in all_fams]
    prts = [os.path.join(aldir, f"{dname}-current.{num}.prt") for num in all_fams]
    assert len(gens + prts) == 12
    assert set(gens + prts) == set(files_todo)


def test_check_extract_prtfile():
    """
    Test that when there is only the prt files for 2 families in aldir (no gen file), and mafft
    file, prt2nuc file for a family + concatenated file, it removes all these files,
    and returns all prt and gen files for all families
    """
    all_fams = [1, 6, 10, 11, 13, 14]
    aldir = os.path.join(GENEPATH, "Aldir-notempty")
    dname = "TESTcheckExtract"
    # Create prt files for families 6 and 13
    # Create mafft and prt2nuc files for family 6
    # Create concatenated file
    os.makedirs(aldir)
    prt1 = os.path.join(aldir, f"{dname}-current.6.prt")
    prt2 = os.path.join(aldir, "{}-current.13.prt".format(dname))
    mafft = os.path.join(aldir, "{}-mafft-align.6.aln".format(dname))
    btr = os.path.join(aldir, "{}-mafft-prt2nuc.6.aln".format(dname))
    concat = os.path.join(aldir, "{}-complete.cat.aln".format(dname))
    for outf in [prt1, prt2, mafft, btr, concat]:
        open(outf, "w").close()
    # Run check extract
    files_todo = gseq.check_existing_extract(all_fams, aldir, dname)
    assert len(files_todo) == 12
    gens = [os.path.join(aldir, "{}-current.{}.gen").format(dname, num) for num in all_fams]
    prts = [os.path.join(aldir, "{}-current.{}.prt").format(dname, num) for num in all_fams]
    assert len(gens + prts) == len(files_todo)
    assert set(gens + prts) == set(files_todo)
    for outf in [prt1, prt2, mafft, btr, concat]:
        assert not os.path.isfile(outf)


def test_check_extract_files1genome():
    """
    Test that when there are already prt, gen, mafft and prt2nuc files for a family, it keeps them,
    and returns all prt and gen files for all other families (except the one already existing).
    + if concatenation file exists, it is removed (as we will realign all families except 1).
    """
    all_fams = [1, 6, 10, 11, 13, 14]
    aldir = os.path.join(GENEPATH, "Aldir-notempty")
    dname = "TESTcheckExtract"
    # Create prt, gen, mafft and prt2nuc files for family 6 + concatenated file
    os.makedirs(aldir)
    prt = os.path.join(aldir, "{}-current.6.prt".format(dname))
    gen = os.path.join(aldir, "{}-current.6.gen".format(dname))
    mafft = os.path.join(aldir, "{}-mafft-align.6.aln".format(dname))
    btr = os.path.join(aldir, "{}-mafft-prt2nuc.6.aln".format(dname))
    concat = os.path.join(aldir, "{}-complete.cat.aln".format(dname))
    for outf in [prt, gen, mafft, btr, concat]:
        open(outf, "w").close()
    # Run check extract
    files_todo = gseq.check_existing_extract(all_fams, aldir, dname)
    assert len(files_todo) == 10
    fams_todo = [1, 10, 11, 13, 14]
    gens = [os.path.join(aldir, f"{dname}-current.{num}.gen") for num in fams_todo]
    prts = [os.path.join(aldir, f"{dname}-current.{num}.prt") for num in fams_todo]
    assert len(gens + prts) == len(files_todo)
    assert set(gens + prts) == set(files_todo)
    for outf in [prt, gen, mafft, btr]:
        assert os.path.isfile(outf)
    assert not os.path.isfile(concat)


def test_check_extract_allexist():
    """
    Test that when there are already all prt and gen files for all families,
    it keeps them, and returns an empty list of files to extract.
    If concatenation, mafft and btr files exist, they are also kept.
    """
    all_fams = [1, 6, 10, 11, 13, 14]
    aldir = os.path.join(GENEPATH, "Aldir-notempty")
    dname = "TESTcheckExtract"
    # Create prt and gen files for all families
    os.makedirs(aldir)
    gens = [os.path.join(aldir, "{}-current.{}.gen").format(dname, num) for num in all_fams]
    prts = [os.path.join(aldir, "{}-current.{}.prt").format(dname, num) for num in all_fams]
    mafft = os.path.join(aldir, f"{dname}-mafft-align.6.aln")
    btr = os.path.join(aldir, f"{dname}-mafft-prt2nuc.6.aln")
    concat = os.path.join(aldir, f"{dname}-complete.cat.aln")
    for outf in gens + prts + [mafft, btr, concat]:
        open(outf, "w").close()
    # Run check extract
    files_todo = gseq.check_existing_extract(all_fams, aldir, dname)
    assert files_todo == []
    for outf in gens + prts + [mafft, btr, concat]:
        assert os.path.isfile(outf)
    shutil.rmtree(aldir)


def test_get_all_seqs(caplog):
    """
    Test that when giving a list of family numbers, and output directories are empty,
    it extracts all expected proteins and genes.
    => Default. empty output, give database and 2 families to extract and getentry files
    exist in Listdir
    """
    caplog.set_level(logging.DEBUG)
    all_genomes = ["GEN2.1017.00001", "GEN4.1111.00001", "GENO.1017.00001", "GENO.1216.00002"]
    dname = "TESTgetAllSeq"
    listdir = os.path.join(GENEPATH, "Listdir")
    aldir = os.path.join(GENEPATH, "Align")
    all_fams = [1, 6]
    quiet = False
    # Create listdir and aldir and put all getentry files in listdir
    os.makedirs(listdir)
    os.makedirs(aldir)
    ref_listdir = os.path.join(TESTPATH, "test_listdir")
    ref_aldir = os.path.join(EXPPATH, "exp_aldir")
    for gen in all_genomes:
        genome_gen = os.path.join(ref_listdir, f"getentry-gen_{gen}")
        genome_prt = os.path.join(ref_listdir, f"getentry-prt_{gen}")
        gen_out = os.path.join(listdir, f"{dname}-getEntry_gen_{gen}.txt")
        prt_out = os.path.join(listdir, f"{dname}-getEntry_prt_{gen}.txt")
        shutil.copyfile(genome_gen, gen_out)
        shutil.copyfile(genome_prt, prt_out)
    gseq.get_all_seqs(all_genomes, dname, DBPATH, listdir, aldir, all_fams, quiet)

    # For each family, check that prt and gen files exist, and their content
    for fam in all_fams:
        fam_prt = os.path.join(aldir, f"{dname}-current.{fam}.prt")
        assert os.path.isfile(fam_prt)
        exp_fam_prt = os.path.join(ref_aldir, f"current.{fam}.prt")
        assert tutil.compare_file_content(fam_prt, exp_fam_prt)
        fam_gen = os.path.join(aldir, f"{dname}-current.{fam}.gen")
        assert os.path.isfile(fam_gen)
        exp_fam_gen = os.path.join(ref_aldir, f"current.{fam}.gen")
        assert tutil.compare_file_content(fam_gen, exp_fam_gen)

    # Check logs
    assert "Extracting proteins and genes from all genomes" in caplog.text
    for gen in all_genomes:
        assert f"Extracting proteins and genes from {gen}" in caplog.text


def test_get_all_seqs_prt6(caplog):
    """
    Test that when giving a list of family numbers, and output directories contain only a prt
    file for 1 family, it removes this prt file and it extracts all expected proteins and genes.
    => Aldir with prt file for fam 6. Others as default
    """
    caplog.set_level(logging.DEBUG)
    all_genomes = ["GEN2.1017.00001", "GEN4.1111.00001", "GENO.1017.00001", "GENO.1216.00002"]
    dname = "TESTgetAllSeq"
    listdir = os.path.join(GENEPATH, "Listdir")
    aldir = os.path.join(GENEPATH, "Align")
    all_fams = [1, 6]
    quiet = False
    # Create listdir and aldir and put all getentry files in listdir
    os.makedirs(listdir)
    os.makedirs(aldir)
    ref_listdir = os.path.join(TESTPATH, "test_listdir")
    ref_aldir = os.path.join(EXPPATH, "exp_aldir")
    prt6 = os.path.join(aldir, "{}-current.6.prt".format(dname))
    # Create empty file for prt of family 6
    open(prt6, "w").close()
    for gen in all_genomes:
        genome_gen = os.path.join(ref_listdir, "getentry-gen_{}".format(gen))
        genome_prt = os.path.join(ref_listdir, "getentry-prt_{}".format(gen))
        gen_out = os.path.join(listdir, "{}-getEntry_gen_{}.txt".format(dname, gen))
        prt_out = os.path.join(listdir, "{}-getEntry_prt_{}.txt".format(dname, gen))
        shutil.copyfile(genome_gen, gen_out)
        shutil.copyfile(genome_prt, prt_out)
    assert os.path.isfile(prt6)
    gseq.get_all_seqs(all_genomes, dname, DBPATH, listdir, aldir, all_fams, quiet)
    # For each family, check that prt and gen files exist, and their content
    for fam in all_fams:
        fam_prt = os.path.join(aldir, "{}-current.{}.prt".format(dname, fam))
        assert os.path.isfile(fam_prt)
        exp_fam_prt = os.path.join(ref_aldir, "current.{}.prt".format(fam))
        assert tutil.compare_file_content(fam_prt, exp_fam_prt)
        fam_gen = os.path.join(aldir, "{}-current.{}.gen".format(dname, fam))
        assert os.path.isfile(fam_gen)
        exp_fam_gen = os.path.join(ref_aldir, "current.{}.gen".format(fam))
        assert tutil.compare_file_content(fam_gen, exp_fam_gen)
    # Check logs
    assert "Extracting proteins and genes from all genomes" in caplog.text
    for gen in all_genomes:
        assert "Extracting proteins and genes from {}".format(gen) in caplog.text


def test_get_all_seqs_prtgen6(caplog):
    """
    Test that when giving a list of family numbers, and output directories contain a prt and a gen
    file for 1 family, it extracts all expected proteins and genes for other families, but keeps
    the current file for family already having prt and gen
    + add mafft and prt2nuc files for this family, and check that they are not removed
    + add concatenate file, and check that it is removed
    => prt and gen files in Aldir for fam 6. Others as default
    """
    caplog.set_level(logging.DEBUG)
    all_genomes = ["GEN2.1017.00001", "GEN4.1111.00001", "GENO.1017.00001", "GENO.1216.00002"]
    dname = "TESTgetAllSeq"
    listdir = os.path.join(GENEPATH, "Listdir")
    aldir = os.path.join(GENEPATH, "Align")
    all_fams = [1, 6]
    quiet = False
    # Create listdir and aldir and put all getentry files in listdir
    os.makedirs(listdir)
    os.makedirs(aldir)
    ref_listdir = os.path.join(TESTPATH, "test_listdir")
    ref_aldir = os.path.join(EXPPATH, "exp_aldir")
    # Create empty files for prt, gen, mafft and prt2nuc files of family 6
    prt6 = os.path.join(aldir, "{}-current.6.prt".format(dname))
    gen6 = os.path.join(aldir, "{}-current.6.gen".format(dname))
    mafft6 = os.path.join(aldir, "{}-mafft-align.6.aln".format(dname))
    prt2nuc6 = os.path.join(aldir, "{}-mafft-prt2nuc.6.aln".format(dname))
    # Add concatenate file
    concat = os.path.join(aldir, "{}-complete.cat.aln".format(dname))
    for outf in [prt6, gen6, mafft6, prt2nuc6, concat]:
        open(outf, "w").close()
    for gen in all_genomes:
        genome_gen = os.path.join(ref_listdir, "getentry-gen_{}".format(gen))
        genome_prt = os.path.join(ref_listdir, "getentry-prt_{}".format(gen))
        gen_out = os.path.join(listdir, "{}-getEntry_gen_{}.txt".format(dname, gen))
        prt_out = os.path.join(listdir, "{}-getEntry_prt_{}.txt".format(dname, gen))
        shutil.copyfile(genome_gen, gen_out)
        shutil.copyfile(genome_prt, prt_out)
    gseq.get_all_seqs(all_genomes, dname, DBPATH, listdir, aldir, all_fams, quiet)
    # For family 1, check that prt and gen files exist and are as expected
    fam_prt = os.path.join(aldir, "{}-current.1.prt".format(dname))
    assert os.path.isfile(fam_prt)
    exp_fam_prt = os.path.join(ref_aldir, "current.1.prt")
    assert tutil.compare_file_content(fam_prt, exp_fam_prt)
    fam_gen = os.path.join(aldir, "{}-current.1.gen".format(dname))
    assert os.path.isfile(fam_gen)
    exp_fam_gen = os.path.join(ref_aldir, "current.1.gen")
    assert tutil.compare_file_content(fam_gen, exp_fam_gen)
    # For family 6 , check that all filesare present and empty
    for outf in [prt6, gen6, mafft6, prt2nuc6]:
        assert os.path.isfile(outf)
        with open(outf, "r") as out:
            assert out.readlines() == []
    # Check that concat file was removed
    assert not os.path.isfile(concat)
    # Check logs
    assert "Extracting proteins and genes from all genomes" in caplog.text
    for gen in all_genomes:
        assert "Extracting proteins and genes from {}".format(gen) in caplog.text


def test_get_all_seqs_allexist(caplog):
    """
    Test that when giving a list of family numbers, and all prt and gen extraction files for this
    family already exist in Aldir, it does not change them, and returns an warning message saying
    that all prt and gen are already extracted.
    + concat file present, and not removed
    """
    caplog.set_level(logging.DEBUG)
    all_genomes = ["GEN2.1017.00001", "GEN4.1111.00001", "GENO.1017.00001", "GENO.1216.00002"]
    dname = "TESTgetAllSeq"
    listdir = os.path.join(GENEPATH, "Listdir")
    aldir = os.path.join(GENEPATH, "Align")
    all_fams = [1, 6]
    quiet = False
    # Create listdir and aldir and put all getentry files in listdir
    os.makedirs(listdir)
    os.makedirs(aldir)
    ref_listdir = os.path.join(TESTPATH, "test_listdir")
    ref_aldir = os.path.join(EXPPATH, "exp_aldir")
    # Create empty files for prt, gen, mafft and prt2nuc files of family 6
    created_files = []
    for fam in all_fams:
        prt = os.path.join(aldir, "{}-current.{}.prt".format(dname, fam))
        gen = os.path.join(aldir, "{}-current.{}.gen".format(dname, fam))
        mafft = os.path.join(aldir, "{}-mafft-align.{}.aln".format(dname, fam))
        prt2nuc = os.path.join(aldir, "{}-mafft-prt2nuc.{}.aln".format(dname, fam))
        created_files += [prt, gen, mafft, prt2nuc]
        for outf in [prt, gen, mafft, prt2nuc]:
            open(outf, "w").close()
    # Add concatenate file
    concat = os.path.join(aldir, "{}-complete.cat.aln".format(dname))
    open(concat, "w").close()
    for gen in all_genomes:
        genome_gen = os.path.join(ref_listdir, "getentry-gen_{}".format(gen))
        genome_prt = os.path.join(ref_listdir, "getentry-prt_{}".format(gen))
        gen_out = os.path.join(listdir, "{}-getEntry_gen_{}.txt".format(dname, gen))
        prt_out = os.path.join(listdir, "{}-getEntry_prt_{}.txt".format(dname, gen))
        shutil.copyfile(genome_gen, gen_out)
        shutil.copyfile(genome_prt, prt_out)
    gseq.get_all_seqs(all_genomes, dname, DBPATH, listdir, aldir, all_fams, quiet)
    # check that all files are present and empty
    for outf in created_files:
        assert os.path.isfile(outf)
        with open(outf, "r") as out:
            assert out.readlines() == []
    # Check that concat file was removed
    assert os.path.isfile(concat)
    # Check logs
    assert ("All extraction files already existing (see detailed log for "
            "more information)") in caplog.text
    assert ("All prt and gene files for all families already exist. The program "
            "will use them for the next step. If you want to re-extract a given "
            "family, remove its prt and gen extraction files. If you want to "
            "re-extract all families, use option -F (or --force).") in caplog.text
